# Orca
#
# Copyright 2004-2009 Sun Microsystems Inc.
# Copyright 2011-2025 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-public-methods

"""The default Script for presenting information to the user."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from orca import (
    action_presenter,
    ax_event_synthesizer,
    braille,
    braille_presenter,
    bypass_mode_manager,
    caret_navigator,
    chat_presenter,
    clipboard,
    cmdnames,
    command_manager,
    debug,
    debugging_tools_manager,
    document_presenter,
    event_manager,
    flat_review_finder,
    flat_review_presenter,
    focus_manager,
    guilabels,
    input_event,
    input_event_manager,
    keybindings,
    learn_mode_presenter,
    live_region_presenter,
    messages,
    mouse_review,
    notification_presenter,
    object_navigator,
    orca,
    orca_gui_prefs,
    presentation_manager,
    profile_manager,
    say_all_presenter,
    script,
    script_manager,
    sleep_mode_manager,
    speech_manager,
    speech_presenter,
    spellcheck_presenter,
    structural_navigator,
    system_information_presenter,
    table_navigator,
    typing_echo_presenter,
    where_am_i_presenter,
)
from orca.ax_document import AXDocument
from orca.ax_object import AXObject
from orca.ax_selection import AXSelection
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities
from orca.ax_utilities_event import TextEventReason
from orca.ax_utilities_text import TextUnit

if TYPE_CHECKING:
    from collections.abc import Callable

    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(script.Script):
    """The default Script for presenting information to the user."""

    _commands_initialized: bool = False

    def _get_all_extensions(self) -> list[tuple[Callable, str]]:
        """Returns (extension_getter, localized_name) for each extension."""

        return [
            (braille_presenter.get_presenter, guilabels.BRAILLE),
            (notification_presenter.get_presenter, guilabels.KB_GROUP_NOTIFICATIONS),
            (clipboard.get_presenter, guilabels.KB_GROUP_CLIPBOARD),
            (command_manager.get_manager, guilabels.KB_GROUP_DEFAULT),
            (say_all_presenter.get_presenter, guilabels.KB_GROUP_DEFAULT),
            (typing_echo_presenter.get_presenter, guilabels.KB_GROUP_DEFAULT),
            (speech_manager.get_manager, guilabels.KB_GROUP_SPEECH_VERBOSITY),
            (speech_presenter.get_presenter, guilabels.KB_GROUP_SPEECH_VERBOSITY),
            (bypass_mode_manager.get_manager, guilabels.KB_GROUP_DEFAULT),
            (sleep_mode_manager.get_manager, guilabels.KB_GROUP_SLEEP_MODE),
            (system_information_presenter.get_presenter, guilabels.KB_GROUP_SYSTEM_INFORMATION),
            (object_navigator.get_navigator, guilabels.KB_GROUP_OBJECT_NAVIGATION),
            (caret_navigator.get_navigator, guilabels.KB_GROUP_CARET_NAVIGATION),
            (structural_navigator.get_navigator, guilabels.KB_GROUP_STRUCTURAL_NAVIGATION),
            (table_navigator.get_navigator, guilabels.KB_GROUP_TABLE_NAVIGATION),
            (document_presenter.get_presenter, guilabels.KB_GROUP_DOCUMENTS),
            (live_region_presenter.get_presenter, guilabels.KB_GROUP_LIVE_REGIONS),
            (learn_mode_presenter.get_presenter, guilabels.KB_GROUP_LEARN_MODE),
            (mouse_review.get_reviewer, guilabels.KB_GROUP_MOUSE_REVIEW),
            (action_presenter.get_presenter, guilabels.KB_GROUP_ACTIONS),
            (flat_review_presenter.get_presenter, guilabels.KB_GROUP_FLAT_REVIEW),
            (flat_review_finder.get_finder, guilabels.KB_GROUP_FIND),
            (where_am_i_presenter.get_presenter, guilabels.KB_GROUP_WHERE_AM_I),
            (debugging_tools_manager.get_manager, guilabels.KB_GROUP_DEBUGGING_TOOLS),
            (chat_presenter.get_presenter, guilabels.KB_GROUP_CHAT),
            (profile_manager.get_manager, guilabels.GENERAL_PROFILES),
        ]

    def set_up_commands(self) -> None:
        """Sets up commands with CommandManager."""

        tokens = ["DEFAULT: Setting up commands for", self.app]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if Script._commands_initialized:
            msg = "DEFAULT: Commands already initialized."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        Script._commands_initialized = True

        manager = command_manager.get_manager()
        group_label = guilabels.KB_GROUP_DEFAULT

        kb_kp_divide_orca = keybindings.KeyBinding("KP_Divide", keybindings.ORCA_MODIFIER_MASK)
        kb_kp_divide = keybindings.KeyBinding("KP_Divide", keybindings.NO_MODIFIER_MASK)
        kb_kp_multiply = keybindings.KeyBinding("KP_Multiply", keybindings.NO_MODIFIER_MASK)
        kb_9_orca = keybindings.KeyBinding("9", keybindings.ORCA_MODIFIER_MASK)
        kb_7_orca = keybindings.KeyBinding("7", keybindings.ORCA_MODIFIER_MASK)
        kb_8_orca = keybindings.KeyBinding("8", keybindings.ORCA_MODIFIER_MASK)
        kb_space_orca = keybindings.KeyBinding("space", keybindings.ORCA_MODIFIER_MASK)
        kb_space_ctrl_orca = keybindings.KeyBinding("space", keybindings.ORCA_CTRL_MODIFIER_MASK)

        script_commands: list[
            tuple[
                str,
                Callable[..., bool],
                str,
                keybindings.KeyBinding | None,
                keybindings.KeyBinding | None,
            ]
        ] = [
            (
                "routePointerToItemHandler",
                self.route_pointer_to_item,
                cmdnames.ROUTE_POINTER_TO_ITEM,
                kb_kp_divide_orca,
                kb_9_orca,
            ),
            (
                "leftClickReviewItemHandler",
                self.left_click_item,
                cmdnames.LEFT_CLICK_REVIEW_ITEM,
                kb_kp_divide,
                kb_7_orca,
            ),
            (
                "rightClickReviewItemHandler",
                self.right_click_item,
                cmdnames.RIGHT_CLICK_REVIEW_ITEM,
                kb_kp_multiply,
                kb_8_orca,
            ),
            (
                "panBrailleLeftHandler",
                self.pan_braille_left,
                cmdnames.PAN_BRAILLE_LEFT,
                None,
                None,
            ),
            (
                "panBrailleRightHandler",
                self.pan_braille_right,
                cmdnames.PAN_BRAILLE_RIGHT,
                None,
                None,
            ),
            (
                "contractedBrailleHandler",
                self.set_contracted_braille,
                cmdnames.SET_CONTRACTED_BRAILLE,
                None,
                None,
            ),
            ("shutdownHandler", self.quit_orca, cmdnames.QUIT_ORCA, None, None),
            (
                "preferencesSettingsHandler",
                self.show_preferences_gui,
                cmdnames.SHOW_PREFERENCES_GUI,
                kb_space_orca,
                kb_space_orca,
            ),
            (
                "appPreferencesSettingsHandler",
                self.show_app_preferences_gui,
                cmdnames.SHOW_APP_PREFERENCES_GUI,
                kb_space_ctrl_orca,
                kb_space_ctrl_orca,
            ),
        ]

        for (
            name,
            function,
            description,
            desktop_kb,
            laptop_kb,
        ) in script_commands:
            manager.add_command(
                command_manager.KeyboardCommand(
                    name,
                    function,
                    group_label,
                    description,
                    desktop_keybinding=desktop_kb,
                    laptop_keybinding=laptop_kb,
                ),
            )

        braille_bindings: dict[str, tuple[int, ...]] = {}
        left_keys = tuple(
            key
            for key in (
                braille.BRLAPI_KEY_CMD_HWINLT,
                braille.BRLAPI_KEY_CMD_FWINLT,
                braille.BRLAPI_KEY_CMD_FWINLTSKIP,
            )
            if key is not None
        )
        right_keys = tuple(
            key
            for key in (
                braille.BRLAPI_KEY_CMD_HWINRT,
                braille.BRLAPI_KEY_CMD_FWINRT,
                braille.BRLAPI_KEY_CMD_FWINRTSKIP,
            )
            if key is not None
        )
        if left_keys:
            braille_bindings["panBrailleLeftHandler"] = left_keys
        if right_keys:
            braille_bindings["panBrailleRightHandler"] = right_keys

        single_key_bindings: list[tuple[str, int | None]] = [
            ("goBrailleHomeHandler", braille.BRLAPI_KEY_CMD_HOME),
            ("contractedBrailleHandler", braille.BRLAPI_KEY_CMD_SIXDOTS),
            ("processRoutingKeyHandler", braille.BRLAPI_KEY_CMD_ROUTE),
            ("processBrailleCutBeginHandler", braille.BRLAPI_KEY_CMD_CUTBEGIN),
            ("processBrailleCutLineHandler", braille.BRLAPI_KEY_CMD_CUTLINE),
        ]
        for handler_name, key in single_key_bindings:
            if key is not None:
                braille_bindings[handler_name] = (key,)
        if not braille_bindings:
            msg = "DEFAULT: Braille bindings unavailable."
            debug.print_message(debug.LEVEL_INFO, msg, True)

        braille_commands: list[tuple[str, Callable[..., bool], str, bool]] = [
            ("panBrailleLeftHandler", self.pan_braille_left, cmdnames.PAN_BRAILLE_LEFT, True),
            ("panBrailleRightHandler", self.pan_braille_right, cmdnames.PAN_BRAILLE_RIGHT, True),
            (
                "contractedBrailleHandler",
                self.set_contracted_braille,
                cmdnames.SET_CONTRACTED_BRAILLE,
                True,
            ),
            ("goBrailleHomeHandler", self.go_braille_home, cmdnames.GO_BRAILLE_HOME, False),
            (
                "processRoutingKeyHandler",
                self.process_routing_key,
                cmdnames.PROCESS_ROUTING_KEY,
                False,
            ),
            (
                "processBrailleCutBeginHandler",
                self.process_braille_cut_begin,
                cmdnames.PROCESS_BRAILLE_CUT_BEGIN,
                False,
            ),
            (
                "processBrailleCutLineHandler",
                self.process_braille_cut_line,
                cmdnames.PROCESS_BRAILLE_CUT_LINE,
                False,
            ),
        ]

        for name, function, description, executes_in_learn_mode in braille_commands:
            bb = braille_bindings.get(name, ())
            manager.add_command(
                command_manager.BrailleCommand(
                    name,
                    function,
                    group_label,
                    description,
                    braille_bindings=bb,
                    executes_in_learn_mode=executes_in_learn_mode,
                ),
            )

        for extension_getter, _localized_name in self._get_all_extensions():
            extension_getter().set_up_commands()

        all_braille_keys: set[int] = set()
        for cmd in manager.get_all_braille_commands():
            all_braille_keys.update(cmd.get_braille_bindings())
        braille.setup_key_ranges(all_braille_keys)

        cmd_count = len(command_manager.get_manager().get_all_keyboard_commands())
        msg = f"DEFAULT: Commands set up: {cmd_count} keyboard commands"
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def register_event_listeners(self) -> None:
        """Registers for listeners needed by this script."""

        event_manager.get_manager().register_script_listeners(self)

    def deregister_event_listeners(self) -> None:
        """De-registers the listeners needed by this script."""

        event_manager.get_manager().deregister_script_listeners(self)

    def _get_queued_event(
        self,
        event_type: str,
        detail1: int | None = None,
        detail2: int | None = None,
        any_data=None,
    ) -> Atspi.Event | None:
        cached_event = self.event_cache.get(event_type, [None, 0])[0]
        if not cached_event:
            tokens = ["SCRIPT: No queued event of type", event_type]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if detail1 is not None and detail1 != cached_event.detail1:
            tokens = [
                "SCRIPT: Queued event's detail1 (",
                str(cached_event.detail1),
                ") doesn't match",
                str(detail1),
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if detail2 is not None and detail2 != cached_event.detail2:
            tokens = [
                "SCRIPT: Queued event's detail2 (",
                str(cached_event.detail2),
                ") doesn't match",
                str(detail2),
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if any_data is not None and any_data != cached_event.any_data:
            tokens = [
                "SCRIPT: Queued event's any_data (",
                cached_event.any_data,
                ") doesn't match",
                any_data,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        tokens = ["SCRIPT: Found matching queued event:", cached_event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return cached_event

    def locus_of_focus_changed(self, event, old_focus, new_focus):
        """Called when the visual object with focus changes."""

        presentation_manager.get_manager().present_command_announcement()

        if not new_focus:
            return True

        if AXUtilities.is_defunct(new_focus):
            return True

        is_name_change = event and event.type.endswith("accessible-name")
        if old_focus == new_focus and not is_name_change:
            msg = "DEFAULT: old focus == new focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.run_find_command_on:
            if self.run_find_command_on == new_focus:
                self.run_find_command_on = None
                flat_review_finder.get_finder().find(self)
            return True

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().quit()

        if learn_mode_presenter.get_presenter().is_active():
            learn_mode_presenter.get_presenter().quit()

        document_presenter.get_presenter().update_mode_if_needed(self, old_focus, new_focus)

        active_window = self.utilities.top_level_object(new_focus)
        focus_manager.get_manager().set_active_window(active_window)
        if old_focus is None:
            old_focus = active_window

        manager = presentation_manager.get_manager()
        manager.interrupt_if_needed_for_focus_change(old_focus, new_focus, event)
        manager.present_object(self, new_focus, priorObj=old_focus)
        return True

    def activate(self) -> None:
        """Called when this script is activated."""

        tokens = ["DEFAULT: Activating script for", self.app]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not focus_manager.get_manager().is_in_preferences_window():
            speech_manager.get_manager().update_punctuation_level()
            speech_manager.get_manager().update_capitalization_style()
            speech_manager.get_manager().update_synthesizer()

        presenter = document_presenter.get_presenter()
        if presenter.has_state_for_app(self.app):
            presenter.restore_mode_for_script(self)
        else:
            reason = "script activation, no prior state"
            presenter.suspend_navigators(self, False, reason)
            structural_navigator.get_navigator().set_mode(self, self._default_sn_mode)
            caret_navigator.get_navigator().set_enabled_for_script(
                self,
                self._default_caret_navigation_enabled,
            )

        table_navigator.get_navigator().refresh_enabled_state()
        command_manager.get_manager().activate_commands(f"activated {self.name}")

    def deactivate(self) -> None:
        """Called when this script is deactivated."""

        if bypass_mode_manager.get_manager().is_active():
            bypass_mode_manager.get_manager().toggle_enabled(self)

    def update_braille(self, obj: Atspi.Accessible, **args) -> None:
        """Updates the braille display to show obj."""

        if not obj:
            return
        braille_presenter.get_presenter().present_generated_braille(self, obj, **args)

    ########################################################################
    #                                                                      #
    # INPUT EVENT HANDLERS (AKA ORCA COMMANDS)                             #
    #                                                                      #
    ########################################################################

    def show_app_preferences_gui(
        self,
        current_script: Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Shows the app Preferences dialog."""

        current_script = current_script or self
        ui = orca_gui_prefs.OrcaSetupGUI(current_script)
        ui.show_gui()
        return True

    def show_preferences_gui(
        self,
        _script: Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Displays the Preferences dialog."""

        default_script = script_manager.get_manager().get_default_script()
        script_manager.get_manager().set_active_script(default_script, "Global preferences")
        ui = orca_gui_prefs.OrcaSetupGUI(default_script)
        ui.show_gui()
        return True

    def quit_orca(
        self,
        _script: Script | None = None,
        _event: input_event.InputEvent | None = None,
    ) -> bool:
        """Quit Orca."""

        orca.shutdown()
        return True

    def pan_braille_left(
        self,
        current_script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display to the left."""

        if (
            isinstance(event, input_event.KeyboardEvent)
            and not braille_presenter.get_presenter().use_braille()
        ):
            msg = "DEFAULT: panBrailleLeft command requires braille or braille monitor"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        target = current_script or self
        if target is not self:
            return target.pan_braille_left(target, event)

        tokens = ["DEFAULT: pan_braille_left using", target]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return self._pan_braille_left(event)

    def _pan_braille_left(self, event: input_event.InputEvent | None = None) -> bool:
        """Pans the braille display to the left."""

        if flat_review_presenter.get_presenter().is_active():
            return flat_review_presenter.get_presenter().pan_braille_left(self, event)

        presenter = braille_presenter.get_presenter()
        if presenter.pan_left():
            return True

        # Couldn't pan (at edge). For text areas, move caret to get more content.
        focus = focus_manager.get_manager().get_locus_of_focus()
        is_text_area = AXUtilities.is_editable(focus) or AXUtilities.is_terminal(focus)
        if not is_text_area:
            return True

        start_offset = AXText.get_line_at_offset(focus)[1]
        moved_caret = False
        if start_offset > 0:
            moved_caret = AXText.set_caret_offset(focus, start_offset - 1)

        # If we didn't move the caret and we're in a terminal, we
        # jump into flat review to review the text.  See
        # http://bugzilla.gnome.org/show_bug.cgi?id=482294.
        if not moved_caret and AXUtilities.is_terminal(focus):
            flat_review_presenter.get_presenter().go_start_of_line(self, event)
            flat_review_presenter.get_presenter().go_previous_character(self, event)

        return True

    def pan_braille_right(
        self,
        current_script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display to the right."""

        if (
            isinstance(event, input_event.KeyboardEvent)
            and not braille_presenter.get_presenter().use_braille()
        ):
            msg = "DEFAULT: panBrailleRight command requires braille or braille monitor"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        target = current_script or self
        if target is not self:
            return target.pan_braille_right(target, event)

        tokens = ["DEFAULT: pan_braille_right using", target]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return self._pan_braille_right(event)

    def _pan_braille_right(self, event: input_event.InputEvent | None = None) -> bool:
        """Pans the braille display to the right."""

        if flat_review_presenter.get_presenter().is_active():
            return flat_review_presenter.get_presenter().pan_braille_right(self, event)

        presenter = braille_presenter.get_presenter()
        if presenter.pan_right():
            return True

        # Couldn't pan (at edge). For text areas, move caret to get more content.
        focus = focus_manager.get_manager().get_locus_of_focus()
        is_text_area = AXUtilities.is_editable(focus) or AXUtilities.is_terminal(focus)
        if not is_text_area:
            return True

        end_offset = AXText.get_line_at_offset(focus)[2]
        if end_offset < AXText.get_character_count(focus):
            AXText.set_caret_offset(focus, end_offset)

        return True

    def go_braille_home(
        self,
        _script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Returns to the component with focus."""

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().quit()
            return True

        presentation_manager.get_manager().interrupt_presentation()
        return braille.return_to_region_with_focus(event)

    def set_contracted_braille(
        self,
        _script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Toggles contracted braille."""

        braille.toggle_contracted_braille(event)
        return True

    def process_routing_key(
        self,
        _script: Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Processes a cursor routing key."""

        # Don't kill flash here because it will restore the previous contents and
        # then process the routing key. If the contents accept a click action, this
        # would result in clicking on the link instead of clearing the flash message.
        presentation_manager.get_manager().interrupt_presentation(kill_flash=False)
        if event is None:
            return True
        braille.process_routing_key(event)
        return True

    def process_braille_cut_begin(
        self,
        _script: Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Clears the selection and moves the caret offset in the currently
        active text area.
        """

        if event is None:
            return True
        caret_context = braille.get_caret_context(event)
        if caret_context.offset < 0:
            return True

        presentation_manager.get_manager().interrupt_presentation()
        AXUtilities.clear_all_selected_text(caret_context.accessible)
        self.utilities.set_caret_offset(caret_context.accessible, caret_context.offset)
        return True

    def process_braille_cut_line(
        self,
        _script: Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Extends the text selection in the currently active text
        area and also copies the selected text to the system clipboard."""

        if event is None:
            return True
        caret_context = braille.get_caret_context(event)
        if caret_context.offset < 0:
            return True

        presentation_manager.get_manager().interrupt_presentation()
        start_offset = AXUtilities.get_selection_start_offset(caret_context.accessible)
        end_offset = AXUtilities.get_selection_end_offset(caret_context.accessible)
        if start_offset < 0 or end_offset < 0:
            caret_offset = AXText.get_caret_offset(caret_context.accessible)
            start_offset = min(caret_context.offset, caret_offset)
            end_offset = max(caret_context.offset, caret_offset)

        AXUtilities.set_selected_text(caret_context.accessible, start_offset, end_offset)
        text = AXUtilities.get_selected_text(caret_context.accessible)[0]
        clipboard.get_presenter().set_text(text)
        return True

    def route_pointer_to_item(
        self,
        _script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Moves the mouse pointer to the current item."""

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().route_pointer_to_object(self, event)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        if ax_event_synthesizer.get_synthesizer().route_to_character(
            focus,
        ) or ax_event_synthesizer.get_synthesizer().route_to_object(focus):
            presentation_manager.get_manager().present_message(messages.MOUSE_MOVED_SUCCESS)
            return True

        full = messages.LOCATION_NOT_FOUND_FULL
        brief = messages.LOCATION_NOT_FOUND_BRIEF
        presentation_manager.get_manager().present_message(full, brief)
        return False

    def left_click_item(
        self,
        _script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Performs a left mouse button click on the current item."""

        if flat_review_presenter.get_presenter().is_active():
            obj = flat_review_presenter.get_presenter().get_current_object(self, event)
            if ax_event_synthesizer.get_synthesizer().try_all_clickable_actions(obj):
                return True
            return flat_review_presenter.get_presenter().left_click_on_object(self, event)

        focus = focus_manager.get_manager().get_locus_of_focus()
        if ax_event_synthesizer.get_synthesizer().try_all_clickable_actions(focus):
            return True

        if AXText.get_character_count(focus):
            if ax_event_synthesizer.get_synthesizer().click_character(focus, None, 1):
                return True

        if ax_event_synthesizer.get_synthesizer().click_object(focus, 1):
            return True

        full = messages.LOCATION_NOT_FOUND_FULL
        brief = messages.LOCATION_NOT_FOUND_BRIEF
        presentation_manager.get_manager().present_message(full, brief)
        return False

    def right_click_item(
        self,
        _script: Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Performs a right mouse button click on the current item."""

        if flat_review_presenter.get_presenter().is_active():
            obj = flat_review_presenter.get_presenter().get_current_object(self, event)
            if ax_event_synthesizer.get_synthesizer().try_all_right_click_actions(obj):
                return True
            flat_review_presenter.get_presenter().right_click_on_object(self, event)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        if ax_event_synthesizer.get_synthesizer().try_all_right_click_actions(focus):
            return True

        if ax_event_synthesizer.get_synthesizer().click_character(focus, None, 3):
            return True

        if ax_event_synthesizer.get_synthesizer().click_object(focus, 3):
            return True

        full = messages.LOCATION_NOT_FOUND_FULL
        brief = messages.LOCATION_NOT_FOUND_BRIEF
        presentation_manager.get_manager().present_message(full, brief)
        return False

    ########################################################################
    #                                                                      #
    # AT-SPI OBJECT EVENT HANDLERS                                         #
    #                                                                      #
    # Event handlers return bool:                                          #
    # - True: Event was fully handled, no further processing needed        #
    # - False: Event wasn't handled, should be passed to parent handlers   #
    # Default return value is True (event handled by default script)       #
    #                                                                      #
    ########################################################################

    def _on_active_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:active accessibility events."""

        window = event.source
        if AXUtilities.is_dialog_or_alert(window) or AXUtilities.is_frame(window):
            if event.detail1 and not AXUtilities.can_be_active_window(window):
                return True

            source_is_active_window = window == focus_manager.get_manager().get_active_window()
            if source_is_active_window and not event.detail1:
                focus = focus_manager.get_manager().get_locus_of_focus()
                if AXUtilities.find_ancestor_inclusive(focus, AXUtilities.is_menu):
                    msg = "DEFAULT: Ignoring event. In menu."
                    debug.print_message(debug.LEVEL_INFO, msg, True)
                    return True

                msg = "DEFAULT: Event is for active window. Clearing state."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_active_window(None)
                return True

            if not source_is_active_window and event.detail1:
                msg = "DEFAULT: Updating active window."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_active_window(
                    window,
                    set_window_as_focus=True,
                    notify_script=True,
                )

        return True

    def _on_active_descendant_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:active-descendant-changed accessibility events."""

        if AXUtilities.is_presentable_active_descendant_change(event):
            focus_manager.get_manager().set_locus_of_focus(event, event.any_data)

        return True

    def _on_checked_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:checked accessibility events."""

        if AXUtilities.is_presentable_checked_change(event):
            self.present_object(event.source, alreadyFocused=True, interrupt=True)

        return True

    def _on_children_added(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:add accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")
        return True

    def _on_children_removed(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:remove accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")
        return True

    # pylint: disable-next=too-many-return-statements
    def _on_caret_moved(self, event: Atspi.Event) -> bool:
        """Callback for object:text-caret-moved accessibility events."""

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.SEARCH_PRESENTABLE:
            msg = "DEFAULT: Presenting line for search result change"
            contents = self.utilities.get_line_contents_at_offset(event.source, event.detail1)
            presentation_manager.get_manager().speak_contents(contents)
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        manager = focus_manager.get_manager()
        focus = manager.get_locus_of_focus()
        if focus != event.source:
            if not AXUtilities.is_focused(event.source):
                msg = "DEFAULT: Change is from unfocused source that is not the locus of focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            # TODO - JD: See if this can be removed. If it's still needed document why.
            manager.set_locus_of_focus(event, event.source, False)

        obj, offset = manager.get_last_cursor_position()
        if offset == event.detail1 and obj == event.source:
            navigation_reasons = {
                TextEventReason.NAVIGATION_BY_WORD,
                TextEventReason.NAVIGATION_BY_CHARACTER,
            }
            if reason not in navigation_reasons:
                msg = "DEFAULT: Event is for last saved cursor position"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True
            msg = "DEFAULT: Position matches but proceeding due to navigation reason"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().interrupt_presentation()

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().quit()

        offset = AXText.get_caret_offset(event.source)

        # TODO - JD: These need to be harmonized / unified / simplified.
        manager.set_last_cursor_position(event.source, offset)
        self.utilities.set_caret_context(event.source, offset)

        ignore = [
            TextEventReason.CUT,
            TextEventReason.PASTE,
            TextEventReason.REDO,
            TextEventReason.UNDO,
        ]
        if reason in ignore:
            msg = f"DEFAULT: Ignoring event due to reason ({reason})"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        if AXUtilities.has_selected_text(event.source):
            msg = "DEFAULT: Event source has text selections"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.utilities.handle_text_selection_change(event.source)
            return True

        text, _start, _end = AXUtilities.get_cached_selected_text(obj)
        if text and self.utilities.handle_text_selection_change(obj):
            msg = "DEFAULT: Event handled as text selection change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "DEFAULT: Presenting text at new caret position"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._present_caret_moved_event(event, reason=reason)
        return True

    def _on_description_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-description events."""

        if AXUtilities.is_presentable_description_change(event):
            presentation_manager.get_manager().present_message(event.any_data)
        return True

    def _on_document_page_changed(self, event: Atspi.Event) -> bool:
        """Callback for document:page-changed accessibility events."""

        if event.detail1 < 0:
            return True

        if not AXDocument.did_page_change(event.source):
            return True

        presentation_manager.get_manager().present_message(messages.PAGE_NUMBER % event.detail1)
        return True

    def _on_expanded_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:expanded accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "expanded-changed event.")
        if not AXUtilities.is_presentable_expanded_change(event):
            return True

        self.present_object(event.source, alreadyFocused=True, interrupt=True)
        details = AXUtilities.get_details_content(event.source)
        for detail in details:
            presentation_manager.get_manager().speak_message(detail)

        return True

    def _on_indeterminate_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:indeterminate accessibility events."""

        if AXUtilities.is_presentable_indeterminate_change(event):
            self.present_object(event.source, alreadyFocused=True, interrupt=True)

        return True

    def _on_invalid_entry_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:invalid-entry accessibility events."""

        if not AXUtilities.is_presentable_invalid_entry_change(event):
            return True

        if event.detail1:
            msg = self.get_speech_generator().get_error_message(event.source)
        else:
            msg = messages.INVALID_ENTRY_FIXED
        presentation_manager.get_manager().speak_message(msg)
        self.update_braille(event.source)
        return True

    def _on_mouse_button(self, event: Atspi.Event) -> bool:
        """Callback for mouse:button events."""

        input_event_manager.get_manager().process_mouse_button_event(event)
        return True

    def _on_announcement(self, event: Atspi.Event) -> bool:
        """Callback for object:announcement events."""

        if isinstance(event.any_data, str):
            presentation_manager.get_manager().present_message(event.any_data)

        return True

    def _on_name_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-name events."""

        if not AXUtilities.is_presentable_name_change(event):
            return True

        manager = focus_manager.get_manager()
        if event.source == manager.get_locus_of_focus():
            # Force the update so that braille is refreshed.
            manager.set_locus_of_focus(event, event.source, True, True)
            return True

        presentation_manager.get_manager().present_message(event.any_data)
        return True

    def _on_object_attributes_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:attributes-changed accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "object-attributes-changed event.")
        return True

    def _on_pressed_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:pressed accessibility events."""

        if AXUtilities.is_presentable_pressed_change(event):
            self.present_object(event.source, alreadyFocused=True, interrupt=True)

        return True

    def _on_selected_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:selected accessibility events."""

        if not AXUtilities.is_presentable_selected_change(event):
            return True

        if speech_presenter.get_presenter().get_only_speak_displayed_text():
            return True

        announce_state = False
        manager = input_event_manager.get_manager()
        if manager.last_event_was_space():
            announce_state = True
        elif (
            manager.last_event_was_up() or manager.last_event_was_down()
        ) and AXUtilities.is_table_cell(event.source):
            announce_state = AXUtilities.is_selected(event.source)

        if not announce_state:
            return True

        # TODO - JD: Unlike the other state-changed callbacks, it seems unwise
        # to call generate_speech() here because that also will present the
        # expandable state if appropriate for the object type. The generators
        # need to gain some smarts w.r.t. state changes.

        if event.detail1:
            presentation_manager.get_manager().speak_message(messages.TEXT_SELECTED)
        else:
            presentation_manager.get_manager().speak_message(messages.TEXT_UNSELECTED)

        return True

    def _on_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:selection-changed accessibility events."""

        if not AXUtilities.is_presentable_selection_change(event):
            return True

        presentation_manager.get_manager().present_command_announcement()

        if self.utilities.handle_container_selection_change(event.source):
            return True

        if (
            AXUtilities.is_page_tab_list(event.source)
            and flat_review_presenter.get_presenter().is_active()
        ):
            # If a wizard-like notebook page being reviewed changes, we might not get
            # any events to update the locusOfFocus. As a result, subsequent flat
            # review commands will continue to present the stale content.
            # TODO - JD: We can potentially do some automatic reading here.
            flat_review_presenter.get_presenter().quit()

        focus = focus_manager.get_manager().get_locus_of_focus()
        mouse_review_item = mouse_review.get_reviewer().get_current_item()
        child = AXUtilities.get_selected_child_for_focus(
            event.source,
            focus,
            lambda c: c == mouse_review_item or AXUtilities.is_layout_only(c),
        )
        if child is not None:
            focus_manager.get_manager().set_locus_of_focus(event, child)

        return True

    def _on_focused_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:focused accessibility events."""

        if not event.detail1:
            return True

        if not AXUtilities.is_focused(event.source):
            tokens = ["DEFAULT:", event.source, "lacks focused state. Clearing cache."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            AXObject.clear_cache(event.source, reason="Event detail1 does not match state.")
            if not AXUtilities.is_focused(event.source):
                msg = "DEFAULT: Clearing cache did not update state."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

        obj = event.source
        window, dialog = self.utilities.frame_and_dialog(obj)
        if window and not AXUtilities.can_be_active_window(window) and not dialog:
            return True

        if AXObject.get_child_count(obj) and not AXUtilities.is_combo_box(obj):
            selected_children = AXUtilities.selected_children(obj)
            if selected_children:
                obj = selected_children[0]

        focus_manager.get_manager().set_locus_of_focus(event, obj)
        return True

    def _on_showing_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:showing accessibility events."""

        obj = event.source
        if AXUtilities.is_notification(obj):
            if not event.detail1:
                return True

            presentation_manager.get_manager().speak_message(
                self.get_speech_generator().get_localized_role_name(obj),
            )
            msg = self.utilities.get_notification_content(obj)
            presenter = presentation_manager.get_manager()
            presenter.speak_accessible_text(obj, msg)
            presenter.present_braille_message(msg)
            notification_presenter.get_presenter().save_notification(msg)
            return True

        if AXUtilities.is_tool_tip(obj):
            was_f1 = input_event_manager.get_manager().last_event_was_f1()
            if not was_f1 and not mouse_review.get_reviewer().get_present_tooltips():
                return True
            if event.detail1:
                self.present_object(obj, interrupt=True)
                return True

            focus = focus_manager.get_manager().get_locus_of_focus()
            if focus and was_f1:
                obj = focus
                self.present_object(obj, priorObj=event.source, interrupt=True)
                return True

        return True

    def _on_text_attributes_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-attributes-changed accessibility events."""

        if not AXUtilities.is_presentable_text_attributes_change(event):
            return True

        word, start, _end = AXText.get_word_at_offset(event.source)
        if not word.strip():
            word, start, _end = AXText.get_word_at_offset(event.source, start - 1)

        speech_pres = speech_presenter.get_presenter()
        if error := speech_pres.get_error_description(event.source, start, False):
            presentation_manager.get_manager().speak_message(error)

        return True

    def _on_text_deleted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:delete accessibility events."""

        if not AXUtilities.is_presentable_text_deletion(event):
            return True

        reason = AXUtilities.get_text_event_reason(event)
        presentation_manager.get_manager().present_command_announcement()
        self.update_braille(event.source)

        if reason == TextEventReason.SELECTED_TEXT_DELETION:
            msg = "DEFAULT: Deletion is believed to be due to deleting selected text"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().present_message(messages.SELECTION_DELETED)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        text = self.utilities.deleted_text(event)
        selected_text, _start, _end = AXUtilities.get_cached_selected_text(event.source)
        if reason == TextEventReason.DELETE:
            msg = "DEFAULT: Deletion is believed to be due to Delete command"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            text = AXText.get_character_at_offset(event.source)[0]
        elif reason == TextEventReason.BACKSPACE and text != selected_text:
            msg = "DEFAULT: Deletion is believed to be due to BackSpace command"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        else:
            msg = "DEFAULT: Event is not being presented due to lack of cause"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if len(text) == 1:
            presentation_manager.get_manager().speak_character(text)
        else:
            presentation_manager.get_manager().speak_accessible_text(event.source, text)

        return True

    def _on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        if not AXUtilities.is_presentable_text_insertion(event):
            return True

        reason = AXUtilities.get_text_event_reason(event)
        presentation_manager.get_manager().present_command_announcement()
        self.update_braille(event.source)

        if reason == TextEventReason.SELECTED_TEXT_RESTORATION:
            msg = "DEFAULT: Insertion is believed to be due to restoring selected text"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().present_message(messages.SELECTION_RESTORED)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        if reason == TextEventReason.LIVE_REGION_UPDATE:
            msg = "DEFAULT: Event is from live region"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            live_region_presenter.get_presenter().handle_event(self, event)
            return True

        reason_messages: dict[TextEventReason, str] = {
            TextEventReason.PAGE_SWITCH: "due to page switch",
            TextEventReason.PASTE: "due to paste",
            TextEventReason.UNSPECIFIED_COMMAND: "due to command",
            TextEventReason.MOUSE_MIDDLE_BUTTON: "due to middle mouse button",
            TextEventReason.TYPING_ECHOABLE: "echoable",
            TextEventReason.AUTO_INSERTION_PRESENTABLE: "presentable auto text event",
            TextEventReason.SELECTED_TEXT_INSERTION: "also selected",
        }
        silent_reasons = {TextEventReason.PAGE_SWITCH, TextEventReason.PASTE}
        description = reason_messages.get(reason)
        if description is not None:
            msg = f"DEFAULT: Insertion is believed to be {description}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            speak_string = reason not in silent_reasons
        else:
            msg = "DEFAULT: Not speaking inserted string due to lack of cause"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            speak_string = False

        # Because some implementations are broken.
        text = self.utilities.inserted_text(event)
        if speak_string:
            if len(text) == 1:
                presentation_manager.get_manager().speak_character(text)
            else:
                presentation_manager.get_manager().speak_accessible_text(event.source, text)

        if len(text) != 1 or reason not in [
            TextEventReason.TYPING,
            TextEventReason.TYPING_ECHOABLE,
        ]:
            return True

        if AXText.get_selected_ranges(event.source):
            return True

        presenter = typing_echo_presenter.get_presenter()
        if presenter.echo_previous_sentence(event.source):
            return True

        presenter.echo_previous_word(event.source)
        return True

    def _on_text_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-selection-changed accessibility events."""

        if AXUtilities.is_focusable(event.source) and not AXUtilities.is_focused(event.source):
            msg = "DEFAULT: Ignoring event from focusable but unfocused source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        # We won't handle undo here as it can lead to double-presentation.
        # If there is an application for which text-changed events are
        # missing upon undo, handle them in an app or toolkit script.

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.UNKNOWN:
            msg = "DEFAULT: Ignoring event because reason for change is unknown"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.SEARCH_PRESENTABLE:
            msg = "DEFAULT: Presenting line believed to be search match"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.say_line(event.source)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.SEARCH_UNPRESENTABLE:
            msg = "DEFAULT: Ignoring event believed to be unpresentable search results change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason in [TextEventReason.CUT, TextEventReason.BACKSPACE, TextEventReason.DELETE]:
            msg = "DEFAULT: Ignoring event believed to be text removal"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.AUTO_SELECTION:
            msg = "DEFAULT: Ignoring auto-selection believed to be side effect of typing"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True
        if reason == TextEventReason.AUTO_UNSELECTION:
            msg = "DEFAULT: Ignoring auto-unselection believed to be side effect of typing"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            AXUtilities.update_cached_selected_text(event.source)
            return True

        self.utilities.handle_text_selection_change(event.source)
        self.update_braille(event.source)
        return True

    def _on_column_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:column-reordered accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "column-reordered event.")
        if not input_event_manager.get_manager().last_event_was_table_sort():
            return True

        if event.source != AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus()):
            return True

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_COLUMNS)
        return True

    def _on_row_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:row-reordered accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "row-reordered event.")
        if not input_event_manager.get_manager().last_event_was_table_sort():
            return True

        if event.source != AXUtilities.get_table(focus_manager.get_manager().get_locus_of_focus()):
            return True

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_ROWS)
        return True

    def _on_value_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-value accessibility events."""

        if not AXUtilities.is_presentable_value_change(event):
            return True

        manager = focus_manager.get_manager()
        if (
            not AXUtilities.is_progress_bar(event.source)
            and event.source != manager.get_locus_of_focus()
        ):
            msg = "DEFAULT: Source != locusOfFocus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_spin_button(event.source):
            manager.set_last_cursor_position(event.source, AXText.get_caret_offset(event.source))

        if not AXUtilities.is_progress_bar(event.source):
            presentation_manager.get_manager().interrupt_presentation()

        presentation_manager.get_manager().present_object(
            self,
            event.source,
            generate_sound=True,
            alreadyFocused=True,
            isProgressBarUpdate=AXUtilities.is_progress_bar(event.source),
        )
        return True

    def _on_window_activated(self, event: Atspi.Event) -> bool:
        """Callback for window:activate accessibility events."""

        if not AXUtilities.can_be_active_window(event.source):
            return True

        manager = focus_manager.get_manager()
        active_window = manager.get_active_window()
        if event.source == active_window:
            msg = "DEFAULT: Event is for active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            spellcheck_presenter.get_presenter().handle_window_event(event, self)
            return True

        manager.set_active_window(event.source)
        if AXUtilities.is_combo_box_popup(active_window):
            return True

        if AXObject.get_child_count(event.source) == 1:
            child = AXObject.get_child(event.source, 0)
            # Popup menus in Chromium live in a menu bar whose first child is a panel.
            if AXUtilities.is_menu_bar(child):
                child = AXUtilities.get_menu(child)
            if AXUtilities.is_menu(child):
                selected_item = AXSelection.get_selected_child(child, 0)
                if AXUtilities.is_selected(selected_item):
                    child = selected_item
                manager.set_locus_of_focus(event, child)
                return True

        if not spellcheck_presenter.get_presenter().handle_window_event(event, self):
            manager.set_locus_of_focus(event, event.source)
        return True

    def _on_window_deactivated(self, event: Atspi.Event) -> bool:
        """Callback for window:deactivate accessibility events."""

        manager = focus_manager.get_manager()
        focus = manager.get_locus_of_focus()
        if AXUtilities.find_ancestor_inclusive(focus, AXUtilities.is_menu):
            msg = "DEFAULT: Ignoring event. In menu."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if event.source != manager.get_active_window():
            msg = "DEFAULT: Ignoring event. Not for active window"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_combo_box_popup(AXUtilities.find_active_window()):
            msg = "DEFAULT: Ignoring event. Combo box popup is the new active window."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_combo_box_popup(event.source):
            msg = "DEFAULT: Ignoring event. Combo box popup."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().quit()

        if learn_mode_presenter.get_presenter().is_active():
            learn_mode_presenter.get_presenter().quit()

        focus_manager.get_manager().clear_state("Window deactivated")
        script_manager.get_manager().set_active_script(None, "Window deactivated")
        return True

    ########################################################################
    #                                                                      #
    # Methods for presenting content                                       #
    #                                                                      #
    ########################################################################

    def _update_braille_caret_position(self, obj: Atspi.Accessible) -> None:
        """Try to reposition the cursor without having to do a full update."""

        if not braille_presenter.get_presenter().use_braille():
            return

        if braille.try_reposition_cursor(obj):
            return

        self.update_braille(obj)

    def _present_caret_moved_event(
        self,
        event: Atspi.Event,
        obj: Atspi.Accessible | None = None,
        reason: TextEventReason = TextEventReason.UNKNOWN,
    ) -> bool:
        """Presents text at the new position, based on heuristics. Returns True if handled."""

        obj = obj or event.source
        self._update_braille_caret_position(obj)

        # TODO - JD: Make this a TextEventReason. Also handle structural navigation
        # and table navigation here. Technically that's not been necessary because
        # it won't match anything below. But it would be cleaner to cover each cause
        # explicitly.
        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "DEFAULT: Event ignored: Last command was caret nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.SAY_ALL:
            msg = "DEFAULT: Not presenting text because SayAll is active"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        navigation_handlers: dict[TextEventReason, Callable[[Atspi.Accessible], None]] = {
            TextEventReason.NAVIGATION_BY_LINE: self.say_line,
            TextEventReason.NAVIGATION_BY_WORD: self.say_word,
            TextEventReason.NAVIGATION_BY_CHARACTER: self.say_character,
            TextEventReason.NAVIGATION_BY_PAGE: self.say_line,
            TextEventReason.NAVIGATION_TO_LINE_BOUNDARY: self.say_character,
            TextEventReason.NAVIGATION_TO_FILE_BOUNDARY: self.say_line,
        }
        handler = navigation_handlers.get(reason)
        if handler is not None:
            handler(obj)
            return True

        if reason == TextEventReason.MOUSE_PRIMARY_BUTTON:
            text, _start, _end = AXUtilities.get_cached_selected_text(event.source)
            if not text:
                self.say_line(obj)
                return True
        return False

    def say_character(self, obj: Atspi.Accessible) -> None:
        """Speak the character at the caret."""

        context_obj, context_offset = self.utilities.get_caret_context(obj)
        if context_obj == obj:
            offset = context_offset
        else:
            offset = AXText.get_caret_offset(obj)

        if input_event_manager.get_manager().last_event_was_forward_caret_selection():
            offset -= 1

        character, start_offset, end_offset = AXText.get_character_at_offset(obj, offset)
        focus_manager.get_manager().emit_region_changed(
            obj,
            start_offset,
            end_offset,
            focus_manager.CARET_TRACKING,
        )

        presentation_manager.get_manager().speak_character_at_offset(obj, offset, character)
        AXUtilities.set_last_text_unit_spoken(TextUnit.CHAR)

    def say_line(self, obj: Atspi.Accessible, offset: int | None = None) -> None:
        """Speaks the line at the current or specified offset."""

        if offset is None:
            offset = AXText.get_caret_offset(obj)
        else:
            AXText.set_caret_offset(obj, offset)

        line, start_offset = AXText.get_line_at_offset(obj, offset)[0:2]
        if line and line != "\n":
            end_offset = start_offset + len(line)
            focus_manager.get_manager().emit_region_changed(
                obj,
                start_offset,
                end_offset,
                focus_manager.CARET_TRACKING,
            )

        speech_presenter.get_presenter().speak_line(
            self,
            obj,
            start_offset,
            start_offset + len(line),
            line,
        )

        AXUtilities.set_last_text_unit_spoken(TextUnit.LINE)

    def say_phrase(self, obj: Atspi.Accessible, start_offset: int, end_offset: int) -> None:
        """Speaks the substring between start and end offset."""

        phrase = self.utilities.expand_eocs(obj, start_offset, end_offset)
        if not phrase:
            return

        if len(phrase) > 1 or phrase.isalnum():
            focus_manager.get_manager().emit_region_changed(
                obj,
                start_offset,
                end_offset,
                focus_manager.CARET_TRACKING,
            )

        speech_presenter.get_presenter().speak_phrase(self, obj, start_offset, end_offset, phrase)
        AXUtilities.set_last_text_unit_spoken(TextUnit.PHRASE)

    def say_word(self, obj: Atspi.Accessible) -> None:
        """Speaks the word at the caret, taking into account the previous caret position."""

        context_obj, context_offset = self.utilities.get_caret_context(obj)
        if context_obj == obj:
            offset = context_offset
        else:
            offset = AXText.get_caret_offset(obj)

        word, start_offset, end_offset = self.utilities.get_word_at_offset_adjusted_for_navigation(
            obj,
            offset,
        )

        speech_pres = speech_presenter.get_presenter()
        if "\n" in word:
            # Announce when we cross a hard line boundary, based on whether or not indentation and
            # justification should be spoken. This was done to avoid yet another setting in
            # response to some users saying this announcement was too chatty. The idea of using
            # this setting for the decision is that if the user wants indentation and justification
            # announced, they are interested in explicit whitespace information.
            if speech_pres.get_speak_indentation_and_justification():
                presentation_manager.get_manager().speak_character("\n")
            if word.startswith("\n"):
                start_offset += 1
            elif word.endswith("\n"):
                end_offset -= 1
            word = AXText.get_substring(obj, start_offset, end_offset)

        # say_phrase is useful because it handles punctuation verbalization, but we don't want
        # to trigger its whitespace presentation.
        matches = list(re.finditer(r"\S+", word))
        if matches:
            start_offset += matches[0].start()
            end_offset -= len(word) - matches[-1].end()
            word = AXText.get_substring(obj, start_offset, end_offset)

        text = word.replace("\n", "\\n")
        msg = f"DEFAULT: Final word at offset {offset} is '{text}' ({start_offset}-{end_offset})"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if error := speech_presenter.get_presenter().get_error_description(obj, start_offset):
            presentation_manager.get_manager().speak_message(error)

        self.say_phrase(obj, start_offset, end_offset)
        AXUtilities.set_last_text_unit_spoken(TextUnit.WORD)

    def present_object(self, obj: Atspi.Accessible, **args) -> None:
        """Presents the current object."""

        tokens = ["DEFAULT: Presenting object", obj, ". Interrupt:", args.get("interrupt", False)]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        offset = args.get("offset")
        if offset is not None:
            AXText.set_caret_offset(obj, offset)

        speech_only = args.pop("speechonly", False)
        presentation_manager.get_manager().present_object(
            self,
            obj,
            generate_braille=not speech_only,
            **args,
        )



# Orca
#
# Copyright 2014-2026 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-instance-attributes

"""Script-customizable support for application spellcheckers."""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from orca import (
    dbus_service,
    debug,
    focus_manager,
    gsettings_registry,
    guilabels,
    messages,
    object_properties,
    preferences_grid_base,
    presentation_manager,
    speech_presenter,
)
from orca.ax_object import AXObject
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities

if TYPE_CHECKING:
    from .scripts import default


@dataclass
class _SpellCheckWidgets:
    """Widget references for an active spellcheck session."""

    window: Atspi.Accessible
    error_widget: Atspi.Accessible
    suggestions_list: Atspi.Accessible
    change_to_entry: Atspi.Accessible | None
    document: Atspi.Accessible | None


@dataclass
class _PresentationState:
    """Tracking state to avoid duplicate presentations."""

    last_presented_suggestion: Atspi.Accessible | None = None
    error_widget_text: str = ""
    context_line: tuple[str, int, int] = field(default_factory=lambda: ("", -1, -1))
    completion_announced: bool = False

    def reset(self) -> None:
        """Reset presentation state for a new spellcheck session."""
        self.last_presented_suggestion = None
        self.error_widget_text = ""
        self.context_line = ("", -1, -1)
        self.completion_announced = False


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Spellcheck", name="spellcheck")
class SpellCheckPresenter:
    """Singleton presenter for spell check support and preferences."""

    _SCHEMA = "spellcheck"
    KEY_SPELL_ERROR = "spell-error"
    KEY_SPELL_SUGGESTION = "spell-suggestion"
    KEY_PRESENT_CONTEXT = "present-context"

    def _get_setting(self, key: str, default: bool) -> bool:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            "b",
            default=default,
        )

    def __init__(self) -> None:
        self._script: default.Script | None = None
        self._widgets: _SpellCheckWidgets | None = None
        self._state: _PresentationState = _PresentationState()
        self._listener: Atspi.EventListener = Atspi.EventListener.new(self._on_event)

        msg = "SPELLCHECK PRESENTER: Registering D-Bus commands."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        controller = dbus_service.get_remote_controller()
        controller.register_decorated_module("SpellCheckPresenter", self)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPELL_ERROR,
        schema="spellcheck",
        gtype="b",
        default=True,
        summary="Spell misspelled word",
        migration_key="spellcheckSpellError",
    )
    @dbus_service.getter
    def get_spell_error(self) -> bool:
        """Returns whether misspelled word should be spelled."""

        return self._get_setting(self.KEY_SPELL_ERROR, True)

    @dbus_service.setter
    def set_spell_error(self, value: bool) -> bool:
        """Sets whether misspelled word should be spelled."""

        if self.get_spell_error() == value:
            return True

        msg = f"SPELLCHECK PRESENTER: Setting spell error to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SPELL_ERROR, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPELL_SUGGESTION,
        schema="spellcheck",
        gtype="b",
        default=True,
        summary="Spell suggested correction",
        migration_key="spellcheckSpellSuggestion",
    )
    @dbus_service.getter
    def get_spell_suggestion(self) -> bool:
        """Returns whether the suggested correction should be spelled."""

        return self._get_setting(self.KEY_SPELL_SUGGESTION, True)

    @dbus_service.setter
    def set_spell_suggestion(self, value: bool) -> bool:
        """Sets whether the suggested correction should be spelled."""

        if self.get_spell_suggestion() == value:
            return True

        msg = f"SPELLCHECK PRESENTER: Setting spell suggestion to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SPELL_SUGGESTION, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PRESENT_CONTEXT,
        schema="spellcheck",
        gtype="b",
        default=True,
        summary="Present context/surrounding sentence",
        migration_key="spellcheckPresentContext",
    )
    @dbus_service.getter
    def get_present_context(self) -> bool:
        """Returns whether to present the context/surrounding sentence."""

        return self._get_setting(self.KEY_PRESENT_CONTEXT, True)

    @dbus_service.setter
    def set_present_context(self, value: bool) -> bool:
        """Sets whether to present the context/surrounding sentence."""

        if self.get_present_context() == value:
            return True

        msg = f"SPELLCHECK PRESENTER: Setting present context to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_PRESENT_CONTEXT, value
        )
        return True

    def create_preferences_grid(self) -> SpellCheckPreferencesGrid:
        """Create and return the spell check preferences grid."""

        return SpellCheckPreferencesGrid(self)

    def _on_event(self, event: Atspi.Event) -> None:
        """Listener for spellcheck-related events."""

        if self._widgets is None:
            return

        if AXObject.is_dead(self._widgets.window):
            self.deactivate()
            return

        if AXUtilities.get_application(event.source) != AXUtilities.get_application(
            self._widgets.window,
        ):
            return

        debug.print_message(debug.LEVEL_INFO, "\nvvvvv PROCESS SPELL CHECK EVENT vvvvv")
        tokens = ["SPELL CHECK PRESENTER: Event:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not AXUtilities.is_ancestor(event.source, self._widgets.window, inclusive=True):
            debug.print_message(debug.LEVEL_INFO, "^^^^^ PROCESS SPELL CHECK EVENT ^^^^^\n")
            return

        if event.type.startswith("object:property-change:accessible-name"):
            self._handle_name_changed(event)
        elif (
            event.type.startswith("object:state-changed:sensitive")
            and self._widgets.change_to_entry
        ):
            self._handle_change_to_entry_sensitive_changed(event)
        elif event.type.startswith("object:selection-changed") or event.type.startswith(
            "object:active-descendant-changed",
        ):
            self._handle_suggestions_list_change(event)
        debug.print_message(debug.LEVEL_INFO, "^^^^^ PROCESS SPELL CHECK EVENT ^^^^^\n")

    def activate(self, window: Atspi.Accessible) -> bool:
        """Activates spellcheck support."""

        tokens = ["SPELL CHECK PRESENTER: Attempting activation for", window]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        error_widget, suggestions_list, change_to_entry = self._find_spellcheck_widgets(window)
        if error_widget is None or suggestions_list is None:
            tokens = ["SPELL CHECK PRESENTER:", window, "is not spellcheck window"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        document = self._get_document_from_cursor_history(window)
        tokens = ["SPELL CHECK PRESENTER: Document:", document]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._widgets = _SpellCheckWidgets(
            window=window,
            error_widget=error_widget,
            suggestions_list=suggestions_list,
            change_to_entry=change_to_entry,
            document=document,
        )
        self._state.reset()

        msg = "SPELL CHECK PRESENTER: Registering event listeners"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._listener.register("object:property-change:accessible-name")
        self._listener.register("object:state-changed:sensitive")
        self._listener.register("object:selection-changed")
        self._listener.register("object:active-descendant-changed")

        # Silently set locus of focus to the deepest focused child in the dialog
        # to prevent focus claims from document objects from interrupting presentation.
        focused = AXUtilities.get_focused_object(window)
        while focused is not None:
            child = AXUtilities.get_focused_object(focused)
            if child is None or child == focused:
                break
            focused = child
        if focused is not None:
            tokens = ["SPELL CHECK PRESENTER: Setting locus of focus to", focused]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            focus_manager.get_manager().set_locus_of_focus(None, focused, notify_script=False)

        msg = "SPELL CHECK PRESENTER: Activation complete"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return True

    def deactivate(self) -> None:
        """Deactivates spellcheck support."""

        if self._widgets is None:
            return

        msg = "SPELL CHECK PRESENTER: Deregistering event listeners"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._listener.deregister("object:property-change:accessible-name")
        self._listener.deregister("object:state-changed:sensitive")
        self._listener.deregister("object:selection-changed")
        self._listener.deregister("object:active-descendant-changed")

        self._clear_state()

    def _get_document_from_cursor_history(
        self,
        window: Atspi.Accessible,
    ) -> Atspi.Accessible | None:
        """Gets the document object from the focus manager's cursor history."""

        manager = focus_manager.get_manager()

        def is_valid_document(obj: Atspi.Accessible | None) -> bool:
            if obj is None or AXObject.is_dead(obj):
                return False
            if obj == window or AXUtilities.is_ancestor(obj, window):
                return False
            if not AXObject.supports_text(obj):
                return False
            return AXUtilities.is_editable(obj) or bool(
                AXUtilities.find_ancestor(obj, AXUtilities.is_editable),
            )

        last_obj, _ = manager.get_last_cursor_position()
        if is_valid_document(last_obj):
            return last_obj

        penult_obj, _ = manager.get_penultimate_cursor_position()
        if is_valid_document(penult_obj):
            return penult_obj

        return None

    def _get_misspelled_word(self) -> str:
        """Returns the misspelled word."""

        if self._widgets is None:
            return ""

        text = AXText.get_all_text(self._widgets.error_widget) or AXObject.get_name(
            self._widgets.error_widget,
        )
        if not text:
            return ""

        # For most apps, the error widget text/name contains only the misspelled word.
        if len(text.split()) == 1:
            return text

        # For soffice, the error widget contains the full sentence with the misspelled word
        # highlighted in red.
        app = AXUtilities.get_application(self._widgets.window)
        app_name = AXObject.get_name(app).lower().split(".")[-1]
        if app_name == "soffice":
            word = self._find_red_text(self._widgets.error_widget)
            if word and not word.isspace():
                return word.strip(".,;:!?")

        return text

    def _find_red_text(self, obj: Atspi.Accessible) -> str:
        """Finds text with red foreground color (misspelled word indicator in soffice)."""

        length = AXText.get_character_count(obj)
        if length <= 0:
            return ""

        offset = 0
        while offset < length:
            attrs, start, end = AXText.get_text_attributes_at_offset(obj, offset)
            if attrs.get("fg-color") == "255,0,0":
                word = AXText.get_substring(obj, start, end)
                tokens = ["SPELL CHECK PRESENTER: Found red text:", word]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return word
            offset = max(end, offset + 1)

        return ""

    def is_active(self) -> bool:
        """Returns True if spellcheck support is currently being used."""

        return self._widgets is not None

    def is_spell_check_window(self, window: Atspi.Accessible) -> bool:
        """Returns True if window is the window/dialog containing the spellcheck."""

        return self._widgets is not None and window == self._widgets.window

    def _is_complete(self) -> bool:
        """Returns True if we have reason to conclude the check is complete."""

        if self._widgets is None or AXObject.is_dead(self._widgets.window):
            msg = "SPELL CHECK PRESENTER: Window is gone; spellcheck complete"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        AXObject.clear_cache(self._widgets.suggestions_list)
        if not AXUtilities.is_sensitive(self._widgets.suggestions_list):
            msg = "SPELL CHECK PRESENTER: Suggestions list is insensitive; spellcheck complete"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def handle_window_event(self, event: Atspi.Event, script: default.Script) -> bool:
        """Handles window activate/deactivate events. Returns True if spellcheck-related."""

        self._script = script
        if event.type.startswith("window:activate"):
            return self._handle_window_activated(event)

        return False

    def _handle_change_to_entry_sensitive_changed(self, event: Atspi.Event) -> bool:
        """Handles sensitive state change on the change-to entry."""

        tokens = [
            "SPELL CHECK PRESENTER: Handling change-to entry sensitive change for",
            event.source,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self._widgets is None or event.source != self._widgets.change_to_entry:
            tokens = ["SPELL CHECK PRESENTER:", event.source, "is not the change-to entry"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        msg = "SPELL CHECK PRESENTER: Change-to entry sensitivity changed. Checking for completion."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return self._present_completion_message()

    def _handle_name_changed(self, event: Atspi.Event) -> bool:
        """Handles name-changed events that might indicate a new misspelled word."""

        tokens = ["SPELL CHECK PRESENTER: Handling name change for", event.source]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self._widgets is None:
            msg = "SPELL CHECK PRESENTER: Spellcheck is not active."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not AXUtilities.is_ancestor(event.source, self._widgets.window, inclusive=True):
            tokens = ["SPELL CHECK PRESENTER:", event.source, "is not in the spellcheck window"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        msg = "SPELL CHECK PRESENTER: Name changed in spellcheck window; checking for error change"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return self._check_and_present_if_error_changed()

    def _handle_suggestions_list_change(self, event: Atspi.Event) -> bool:
        """Handles selection-changed and active-descendant-changed events for suggestions list."""

        tokens = [
            "SPELL CHECK PRESENTER: Handling suggestions list change for",
            event.source,
            "Event type:",
            event.type,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self._widgets is None or event.source != self._widgets.suggestions_list:
            tokens = ["SPELL CHECK PRESENTER:", event.source, "is not the suggestions list"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        items = AXUtilities.get_selected_children(event.source)
        selected_item = items[0] if len(items) == 1 else None

        if AXUtilities.is_focused(event.source):
            if selected_item is not None:
                assert self._script is not None
                msg = "SPELL CHECK PRESENTER: List is focused; presenting selected suggestion"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                # If focus is newly entering the list, clear last_presented_suggestion
                # so we announce the item even if it was presented during error details.
                current_focus = focus_manager.get_manager().get_locus_of_focus()
                if not AXUtilities.is_ancestor(current_focus, self._widgets.suggestions_list):
                    self._state.last_presented_suggestion = None
                focus_manager.get_manager().set_locus_of_focus(
                    None,
                    selected_item,
                    notify_script=False,
                )
                self._script.update_braille(selected_item)
                self._present_suggestion_list_item()
            return True

        if selected_item == self._state.last_presented_suggestion:
            tokens = [
                "SPELL CHECK PRESENTER: Selection unchanged since last presentation:",
                selected_item,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        msg = "SPELL CHECK PRESENTER: List is not focused; checking for error change"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        # If we have no document, reset last error text to force presentation.
        if self._widgets.document is None:
            self._state.error_widget_text = ""

        self._check_and_present_if_error_changed()
        return True

    def _check_and_present_if_error_changed(self) -> bool:
        """Checks if the misspelled word changed and presents if so."""

        if self._widgets is None or self._state.completion_announced:
            reason = "Not active" if self._widgets is None else "Completion already announced"
            msg = f"SPELL CHECK PRESENTER: {reason}; not checking for error change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._is_complete():
            msg = "SPELL CHECK PRESENTER: Spellcheck is complete; presenting completion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return self._present_completion_message()

        # Check full error widget text first - if it changed, we have a new error
        error_widget_text = AXText.get_all_text(self._widgets.error_widget) or AXObject.get_name(
            self._widgets.error_widget,
        )
        error_text_changed = error_widget_text != self._state.error_widget_text

        current_word = self._get_misspelled_word()
        if not current_word:
            msg = "SPELL CHECK PRESENTER: No current error word"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        # Multi-word content in non-editable widget is not a misspelled word (may be completion msg)
        if len(current_word.split()) > 1 and not AXUtilities.is_editable(
            self._widgets.error_widget,
        ):
            return self._present_completion_message()

        # Check if context line changed (detects same word at different position)
        current_context_line: tuple[str, int, int] = ("", -1, -1)
        if self._widgets.document is not None:
            current_context_line = AXText.get_line_at_offset(self._widgets.document)
        context_changed = current_context_line != self._state.context_line

        msg = (
            f"SPELL CHECK PRESENTER: error_text_changed={error_text_changed} "
            f"context_changed={context_changed}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if not error_text_changed and not context_changed:
            msg = "SPELL CHECK PRESENTER: Error text and context unchanged; not presenting"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "SPELL CHECK PRESENTER: Error changed; presenting"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._state.last_presented_suggestion = None
        self.present_error_details()
        self._state.error_widget_text = error_widget_text
        self._state.context_line = current_context_line
        return True

    def _handle_window_activated(self, event: Atspi.Event) -> bool:
        """Handles window activation."""

        tokens = ["SPELL CHECK PRESENTER: Handling window:activate for", event.source]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not (self.is_spell_check_window(event.source) or self.activate(event.source)):
            msg = "SPELL CHECK PRESENTER: Not a spellcheck window. Deactivating."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.deactivate()
            return False

        msg = "SPELL CHECK PRESENTER: Window activated; checking for error to present"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._state.error_widget_text = ""  # Reset to force presentation
        self._state.context_line = ("", -1, -1)
        self._check_and_present_if_error_changed()
        if self._widgets is not None and self._widgets.change_to_entry is not None:
            assert self._script is not None
            focus_manager.get_manager().set_locus_of_focus(
                None,
                self._widgets.change_to_entry,
                False,
            )
            self._script.update_braille(self._widgets.change_to_entry)
        return True

    def _present_context(self) -> bool:
        """Presents the context/surrounding content of the misspelled word."""

        if self._widgets is None:
            return False

        word = self._get_misspelled_word()
        if not word:
            return False

        if self._script is None:
            return False

        string = self._get_context_string(word)
        if not string:
            return False

        msg = messages.MISSPELLED_WORD_CONTEXT % string.strip()
        presentation_manager.get_manager().speak_message(msg)
        return True

    def _get_context_string(self, word: str) -> str:
        """Returns the context string containing the misspelled word."""

        if self._widgets is None:
            return ""

        error_text = AXText.get_all_text(self._widgets.error_widget)
        if error_text and word in error_text and len(error_text.split()) > 1:
            tokens = ["SPELL CHECK PRESENTER: Using error widget text as context"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return error_text

        if self._widgets.document is None:
            return ""

        string, start, end = AXText.get_line_at_offset(self._widgets.document)
        msg = f"SPELL CHECK PRESENTER: Line at offset {start}-{end}: {string}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if not string or word not in string:
            return ""

        if len(string) > 80:
            sentences = re.split(r"(?:\.|\!|\?)", string)
            if string.count(word) == 1:
                match = list(filter(lambda x: word in x, sentences))
                if match:
                    return match[0].strip()

        return string

    def _present_completion_message(self) -> bool:
        """Presents the message that spellcheck is complete."""

        if self._widgets is None or not self._is_complete():
            return False

        if self._state.completion_announced:
            msg = "SPELL CHECK PRESENTER: Completion already announced"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self._script is None:
            return False

        msg = AXText.get_all_text(self._widgets.error_widget) or AXObject.get_name(
            self._widgets.error_widget,
        )
        manager = presentation_manager.get_manager()
        manager.speak_message(msg)

        # Don't restore previous braille content to prevent accidental clicks on old elements.
        manager.present_braille_message(msg, restore_previous=False)
        self._state.completion_announced = True
        return True

    def present_error_details(
        self,
        detailed: bool = False,
        script: default.Script | None = None,
    ) -> bool:
        """Presents the details of the error."""

        if script is not None:
            self._script = script

        if self._is_complete():
            msg = "SPELL CHECK PRESENTER: Not presenting error details: spellcheck is complete"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._present_mistake(detailed):
            self._present_suggestion(detailed)
            if detailed or self.get_present_context():
                self._present_context()
            return True

        return False

    def _present_mistake(self, detailed: bool = False) -> bool:
        """Presents the misspelled word."""

        if self._widgets is None:
            msg = "SPELL CHECK PRESENTER: Not presenting mistake: spellcheck is not active"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._script is None:
            return False

        word = self._get_misspelled_word()
        if not word:
            msg = "SPELL CHECK PRESENTER: Not presenting mistake: misspelled word not found"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = messages.MISSPELLED_WORD % word
        presentation_manager.get_manager().speak_message(msg)
        if detailed or self.get_spell_error():
            presentation_manager.get_manager().spell_item(word)

        return True

    def _present_suggestion(self, detailed: bool = False) -> bool:
        """Presents the suggested correction."""

        if self._widgets is None:
            msg = "SPELL CHECK PRESENTER: Not presenting suggestion: spellcheck is not active"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not self._widgets.change_to_entry:
            return self._present_suggestion_list_item(detailed, include_label=True)

        if self._script is None:
            return False

        label = AXUtilities.get_displayed_label(self._widgets.change_to_entry) or AXObject.get_name(
            self._widgets.change_to_entry,
        )
        string = AXText.get_substring(self._widgets.change_to_entry, 0, -1)
        msg = f"{label} {string}"
        presentation_manager.get_manager().speak_message(msg)
        if detailed or self.get_spell_suggestion():
            presentation_manager.get_manager().spell_item(string)

        items = AXUtilities.get_selected_children(self._widgets.suggestions_list)
        if len(items) == 1:
            self._state.last_presented_suggestion = items[0]

        return True

    def _present_suggestion_list_item(
        self,
        detailed: bool = False,
        include_label: bool = False,
    ) -> bool:
        """Presents the current item from the suggestions list."""

        if self._widgets is None:
            msg = "SPELL CHECK PRESENTER: Not presenting suggested item: spellcheck is not active"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._script is None:
            return False

        items = AXUtilities.get_selected_children(self._widgets.suggestions_list)
        if len(items) != 1:
            return False

        if items[0] == self._state.last_presented_suggestion:
            return False
        self._state.last_presented_suggestion = items[0]

        if include_label:
            label = AXUtilities.get_displayed_label(
                self._widgets.suggestions_list,
            ) or AXObject.get_name(self._widgets.suggestions_list)
        else:
            label = ""
        string = AXObject.get_name(items[0])

        msg = f"{label} {string}"
        presentation_manager.get_manager().speak_message(msg.strip())
        if detailed or self.get_spell_suggestion():
            presentation_manager.get_manager().spell_item(string)

        if (
            speech_presenter.get_presenter().get_speak_position_in_set()
            and items[0] == focus_manager.get_manager().get_locus_of_focus()
        ):
            index = AXUtilities.get_position_in_set(items[0]) + 1
            total = AXUtilities.get_set_size(items[0])
            msg = object_properties.GROUP_INDEX_SPEECH % {"index": index, "total": total}
            presentation_manager.get_manager().speak_message(msg)

        return True

    def _clear_state(self) -> None:
        """Clears all session state."""
        self._script = None
        self._widgets = None
        self._state.reset()

    def _has_id(self, obj: Atspi.Accessible, target_id: str) -> bool:
        """Returns True if obj has target_id as accessible ID or object attribute 'id'."""

        return (
            AXObject.get_accessible_id(obj) == target_id
            or AXObject.get_attribute(obj, "id") == target_id
        )

    def _could_be_spellcheck_window(self, window: Atspi.Accessible, app_name: str) -> bool:
        """Returns True if window could be the spellcheck window."""

        # What we want. LibreOffice uses this.
        if self._has_id(window, "SpellingDialog"):
            return True

        # Fallback: thunderbird uses frame with tag=body; no ID.
        # Unfortunately, thunderbird does this for the main window with tons of descendants.
        if app_name == "thunderbird":
            if AXObject.get_attribute(window, "tag") != "body":
                return False
            return AXUtilities.get_menu(window) is None

        # Fallback: gedit/pluma uses a modal dialog (gedit) or frame (pluma); no ID.
        if app_name in ("gedit", "pluma"):
            return AXUtilities.is_modal(window)

        return False

    def _could_be_error_widget(self, obj: Atspi.Accessible, app_name: str) -> bool:
        """Returns True if obj could be the error widget."""

        # What we want. Thunderbird uses this.
        if self._has_id(obj, "MisspelledWord"):
            return True

        # Fallback: soffice uses "errorsentence".
        if app_name == "soffice":
            return AXObject.get_accessible_id(obj) == "errorsentence"

        # Fallback: gedit/pluma uses label with single word, no colon.
        if app_name in ("gedit", "pluma") and AXUtilities.is_label(obj):
            name_tokens = AXObject.get_name(obj).split()
            return len(name_tokens) == 1 and ":" not in name_tokens[0]

        return False

    def _could_be_suggestions_list(self, obj: Atspi.Accessible, app_name: str) -> bool:
        """Returns True if obj could be the suggestions list."""

        # What we want. Thunderbird uses "SuggestedList", but the "SuggestionsList" seems nicer,
        # given the label is typically "Suggestions:". But we'll happily accept either.
        if self._has_id(obj, "SuggestedList") or self._has_id(obj, "SuggestionsList"):
            return True

        # Fallback: soffice uses "suggestionslb".
        if app_name == "soffice":
            return AXObject.get_accessible_id(obj) == "suggestionslb"

        # Fallback: gedit/pluma uses a table which manages descendants.
        if app_name in ("gedit", "pluma"):
            return AXUtilities.is_table(obj) and AXUtilities.manages_descendants(obj)

        return False

    def _could_be_change_to_entry(self, obj: Atspi.Accessible, app_name: str) -> bool:
        """Returns True if obj could be the change-to entry."""

        # TODO - JD: What we want (need to try to find consensus).

        # soffice lacks a change-to entry.
        if app_name == "soffice":
            return False

        # Fallback: Thunderbird uses this.
        if self._has_id(obj, "ReplaceWordInput"):
            return True

        # Fallback: gedit/pluma uses labelled single-line editable
        if app_name in ("gedit", "pluma"):
            return AXUtilities.is_single_line_entry(obj)

        return False

    def _find_spellcheck_widgets(
        self,
        window: Atspi.Accessible,
    ) -> tuple[Atspi.Accessible | None, Atspi.Accessible | None, Atspi.Accessible | None]:
        """Finds spellcheck widgets. Returns (error_widget, suggestions_list, change_to_entry)."""

        app = AXUtilities.get_application(window)
        app_name = AXObject.get_name(app).lower().split(".")[-1]

        if not self._could_be_spellcheck_window(window, app_name):
            return None, None, None

        error_widget = AXUtilities.find_descendant(
            window,
            lambda obj: self._could_be_error_widget(obj, app_name),
        )
        suggestions_list = AXUtilities.find_descendant(
            window,
            lambda obj: self._could_be_suggestions_list(obj, app_name),
        )
        change_to_entry = (
            None
            if app_name == "soffice"
            else AXUtilities.find_descendant(
                window,
                lambda obj: self._could_be_change_to_entry(obj, app_name),
            )
        )

        # soffice: errorsentence ID is on panel; get text child if needed
        if error_widget is not None and not AXObject.supports_text(error_widget):
            if text_child := AXUtilities.get_descendant_supporting_text(error_widget):
                error_widget = text_child

        tokens = [
            "SPELL CHECK PRESENTER: Error widget:",
            error_widget,
            "Suggestions list:",
            suggestions_list,
            "Change-to entry:",
            change_to_entry,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return error_widget, suggestions_list, change_to_entry


class SpellCheckPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Spell Check preferences page."""

    _gsettings_schema = "spellcheck"

    def __init__(self, presenter: SpellCheckPresenter) -> None:
        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_SPELL_ERROR,
                getter=presenter.get_spell_error,
                setter=presenter.set_spell_error,
                prefs_key=SpellCheckPresenter.KEY_SPELL_ERROR,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_SPELL_SUGGESTION,
                getter=presenter.get_spell_suggestion,
                setter=presenter.set_spell_suggestion,
                prefs_key=SpellCheckPresenter.KEY_SPELL_SUGGESTION,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_PRESENT_CONTEXT,
                getter=presenter.get_present_context,
                setter=presenter.set_present_context,
                prefs_key=SpellCheckPresenter.KEY_PRESENT_CONTEXT,
            ),
        ]

        super().__init__(
            guilabels.SPELL_CHECK,
            controls,
            info_message=guilabels.SPELL_CHECK_DESCRIPTION,
        )


_presenter: SpellCheckPresenter = SpellCheckPresenter()


def get_presenter() -> SpellCheckPresenter:
    """Returns the Spell Check Presenter"""

    return _presenter



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=unused-argument

"""Functionality for working with speech servers."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from . import gsettings_registry
from .ax_utilities_debugging import AXUtilitiesDebugging

DEFAULT_VOICE: str = "default"
UPPERCASE_VOICE: str = "uppercase"
HYPERLINK_VOICE: str = "hyperlink"
SYSTEM_VOICE: str = "system"


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.PunctuationStyle",
    values={"all": 0, "most": 1, "some": 2, "none": 3},
)
class PunctuationStyle(Enum):
    """Punctuation style enumeration."""

    NONE = 3
    SOME = 2
    MOST = 1
    ALL = 0

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.CapitalizationStyle",
    values={"none": 0, "spell": 1, "icon": 2},
)
class CapitalizationStyle(Enum):
    """Capitalization style enumeration with string values from settings."""

    NONE = "none"
    SPELL = "spell"
    ICON = "icon"

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import ClassVar

    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from . import input_event
    from .acss import ACSS


class VoiceFamily(dict):
    """Holds the family description for a voice."""

    NAME = "name"
    GENDER = "gender"
    LANG = "lang"
    DIALECT = "dialect"
    VARIANT = "variant"

    MALE = "male"
    FEMALE = "female"

    settings: ClassVar[dict[str, str | None]] = {
        NAME: None,
        GENDER: None,
        LANG: None,
        DIALECT: None,
        VARIANT: None,
    }

    def __init__(self, props: dict[str, Any] | None) -> None:
        """Create and initialize VoiceFamily."""
        dict.__init__(self)

        self.update(VoiceFamily.settings)
        if props:
            self.update(props)


class SayAllContext:
    """Contains information about the current state of a Say All operation."""

    PROGRESS = 0
    INTERRUPTED = 1
    COMPLETED = 2

    def __init__(
        self,
        obj: Atspi.Accessible,
        utterance: str,
        start_offset: int = -1,
        end_offset: int = -1,
    ) -> None:
        self.obj = obj
        self.utterance = utterance
        self.start_offset = start_offset
        self.end_offset = end_offset
        self.current_offset = start_offset
        self.current_end_offset = None

    def __str__(self) -> str:
        obj_string = AXUtilitiesDebugging.as_string(self.obj)
        return (
            f"SAY ALL: {obj_string} '{self.utterance}' ({self.start_offset}-{self.end_offset}, "
            f"current: {self.current_offset})"
        )

    def copy(self) -> SayAllContext:
        """Returns a copy of this SayAllContext."""

        new = SayAllContext(self.obj, self.utterance, self.start_offset, self.end_offset)
        new.current_offset = self.current_offset
        new.current_end_offset = self.current_end_offset
        return new

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SayAllContext):
            return False
        return (
            self.start_offset == other.start_offset
            and self.end_offset == other.end_offset
            and self.obj == other.obj
            and self.utterance == other.utterance
        )

    def __hash__(self) -> int:
        return hash((self.start_offset, self.end_offset, self.obj, self.utterance))


class SpeechServer:
    """Provides speech server abstraction."""

    @staticmethod
    def get_factory_name() -> str:
        """Returns a localized name describing this factory."""

        return ""

    @staticmethod
    def get_speech_servers() -> list[Any]:
        """Returns a list of available speech servers."""

        return []

    @staticmethod
    def get_speech_server(_info: Any) -> SpeechServer | None:
        """Gets a given SpeechServer based upon the [name, id] info."""

        return None

    @staticmethod
    def shutdown_active_servers() -> None:
        """Cleans up and shuts down this factory."""

    def get_info(self) -> list[str]:
        """Returns [name, id] of the current speech server."""

        return ["", ""]

    def get_voice_families(self) -> list[VoiceFamily]:
        """Returns a list of all known VoiceFamily instances provided by the server."""

        return []

    def get_voice_family(self) -> VoiceFamily:
        """Returns the current voice family as a VoiceFamily dictionary."""

        return VoiceFamily({})

    def set_voice_family(self, family: VoiceFamily) -> None:
        """Sets the voice family to family VoiceFamily dictionary."""

    def speak_character(
        self,
        character: str,
        acss: ACSS | None = None,
        cap_style: CapitalizationStyle | None = None,
    ) -> None:
        """Speaks character."""

    def speak_key_event(self, event: input_event.KeyboardEvent, acss: ACSS | None = None) -> None:
        """Speaks event."""

    def speak(self, text: str | None = None, acss: ACSS | None = None) -> None:
        """Speaks text using the voice specified by acss."""

    def say_all(
        self,
        utterance_iterator: Iterator[tuple[SayAllContext, ACSS]],
        progress_callback: Callable[[SayAllContext, int], None],
    ) -> None:
        """Iterates through the given utterance_iterator, speaking each utterance."""

    def set_default_voice(self, default_voice: dict[str, Any]) -> None:
        """Sets the default voice ACSS properties for fallback use."""

    def increase_speech_rate(self, step: int = 5) -> None:
        """Increases the speech rate."""

    def decrease_speech_rate(self, step: int = 5) -> None:
        """Decreases the speech rate."""

    def increase_speech_pitch(self, step: float = 0.5) -> None:
        """Increases the speech pitch."""

    def decrease_speech_pitch(self, step: float = 0.5) -> None:
        """Decreases the speech pitch."""

    def increase_speech_volume(self, step: float = 0.5) -> None:
        """Increases the speech volume."""

    def decrease_speech_volume(self, step: float = 0.5) -> None:
        """Decreases the speech volume."""

    def update_capitalization_style(self, style: str) -> None:
        """Updates the capitalization style used by the speech server."""

    def update_punctuation_level(self, level: PunctuationStyle) -> None:
        """Punctuation level changed, inform this speechServer."""

    def stop(self) -> None:
        """Stops ongoing speech and flushes the queue."""

    def shutdown(self) -> None:
        """Shuts down the speech engine."""

    def reset(self) -> None:
        """Resets the speech engine."""

    def clear_cached_voice_properties(self) -> None:
        """Clear cached voice properties to force reapplication on next speech."""

    def get_output_module(self) -> str:
        """Returns the output module associated with this speech server."""

        return ""

    def set_output_module(self, module_id: str) -> None:
        """Sets the output module associated with this speech server."""

    def list_output_modules(self) -> tuple[str, ...]:
        """Return names of available output modules as a tuple of strings."""

        return ()

    def get_voice_families_for_language(
        self,
        language: str,
        dialect: str = "",
        variant: str | None = None,
        maximum: int | None = None,
    ) -> list[tuple[str, str, str | None]]:
        """Returns the families for language available in the current synthesizer."""

        return []

    def _get_language_and_dialect(self, acss_family: dict[str, Any] | None) -> tuple[str, str]:
        """Returns the language and dialect from the ACSS family dictionary."""

        return "", ""

    def get_language_and_dialect(self, acss_family: dict[str, Any] | None) -> tuple[str, str]:
        """Returns the language and dialect from the ACSS family dictionary."""

        return self._get_language_and_dialect(acss_family)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2025 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes

"""Provides the default implementation for flat review for Orca."""

from __future__ import annotations

import re

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import braille, braille_presenter, debug, focus_manager, script_manager, speech_presenter
from .ax_component import AXComponent
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities


class Char:
    """A character's worth of presentable information."""

    def __init__(self, word: Word, start_offset: int, string: str):
        """Creates a new char.

        Arguments:
        - word: the Word instance this belongs to
        - start_offset: the start offset with respect to the accessible object
        - string: the char string
        """

        self._word: Word = word
        self._start_offset: int = start_offset
        self._string: str = string
        self._rect: Atspi.Rect | None = None

    def __str__(self) -> str:
        text = self._string.replace("\n", "\\n")
        rect = self.get_rect()
        rect_string = f"({rect.x}, {rect.y}, {rect.width}, {rect.height})"
        return f"CHAR: '{text}' ({self._start_offset}-{self._start_offset + 1}) rect: {rect_string}"

    def get_string(self) -> str:
        """Returns the string being displayed for this Char."""

        return self._string

    def get_start_offset(self) -> int:
        """Returns the start offset of this Char with respect to the accessible object."""

        return self._start_offset

    def get_rect(self) -> Atspi.Rect:
        """Returns the Atspi.Rect instance representing the extents of this Char."""

        if self._rect is None:
            obj = self._word.get_object()
            self._rect = AXText.get_character_rect(obj, self._start_offset)

        return self._rect


class Word:
    """A single chunk (word or object) of presentable information."""

    def __init__(self, zone: Zone, start_offset: int, string: str):
        """Creates a new Word.

        Arguments:
        - zone: the Zone instance this belongs to
        - start_offset: the start offset with respect to the accessible object
        - string: the word string
        """

        self._zone: Zone = zone
        self._start_offset: int = start_offset
        self._string: str = string
        self._rect: Atspi.Rect | None = None
        self._characters: dict[int, Char] = {}

    def __str__(self) -> str:
        text = self._string.replace("\n", "\\n")
        rect = self.get_rect()
        rect_string = f"({rect.x}, {rect.y}, {rect.width}, {rect.height})"
        return (
            f"WORD: '{text}' ({self._start_offset}-{self._start_offset + len(self._string)}) "
            f"rect: {rect_string}"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Word):
            return False

        if (
            self._zone != other._zone
            or self._start_offset != other._start_offset
            or self._string != other._string
        ):
            return False

        this_rect = self.get_rect()
        other_rect = other.get_rect()
        return (
            this_rect.x == other_rect.x
            and this_rect.y == other_rect.y
            and this_rect.width == other_rect.width
            and this_rect.height == other_rect.height
        )

    def __hash__(self) -> int:
        return hash((self._zone, self._start_offset, self._string))

    def get_characters(self) -> list[Char]:
        """Returns a list of Char instances for this Word."""

        for i, char_str in enumerate(self._string):
            if i not in self._characters:
                start = i + self._start_offset
                self._characters[i] = Char(self, start, char_str)

        return [self._characters[i] for i in range(len(self._string))]

    def get_character_at_index(self, index: int) -> Char | None:
        """Returns the Char at the specified index with respect to this Word."""

        if not 0 <= index < len(self._string):
            return None

        if index not in self._characters:
            char_str = self._string[index]
            start = index + self._start_offset
            self._characters[index] = Char(self, start, char_str)

        return self._characters[index]

    def get_start_offset(self) -> int:
        """Returns the start offset of this Word with respect to the accessible object."""

        return self._start_offset

    def get_relative_offset(self, offset: int) -> int:
        """Returns the char offset with respect to this word or -1."""

        if self._start_offset <= offset < self._start_offset + len(self._string):
            return offset - self._start_offset

        return -1

    def get_string(self) -> str:
        """Returns the string being displayed for this Word."""

        return self._string

    def get_rect(self) -> Atspi.Rect:
        """Returns the Atspi.Rect instance representing the extents of this Word."""

        if self._rect is None:
            self._rect = self._zone.get_word_rect(self._start_offset, self._string)
        return self._rect

    def get_object(self) -> Atspi.Accessible:
        """Returns the accessible object associated with this Word's zone."""

        return self._zone.get_object()


class Zone:
    """Represents text that is a portion of a single horizontal line."""

    WORDS_RE = re.compile(r"(\S+\s*)")

    def __init__(self, obj: Atspi.Accessible, string: str, rect: Atspi.Rect) -> None:
        """Creates a new Zone.

        Arguments:
        - obj: the Accessible associated with this Zone
        - string: the string being displayed for this Zone
        - rect: an Atspi.Rect instance representing the extents of this Zone
        """

        self._obj: Atspi.Accessible = obj
        self._start_offset: int = 0
        self._string: str = string
        self._rect: Atspi.Rect = rect
        self._words: list[Word] = []
        self.line: Line | None = None
        self._braille_region: braille.Region | None = None
        self._word_rect_cache: dict[tuple[int, int], Atspi.Rect] = {}
        self._word_index_map: dict[tuple[int, str], int] = {}  # (start_offset, string) -> index

    def __str__(self) -> str:
        text = self._string.replace("\n", "\\n")
        return f"ZONE: '{text}' {AXObject.get_role_name(self._obj)}"

    def _can_cache_braille_region(self) -> bool:
        """Returns True if we can cache the braille region for this Zone."""

        if AXUtilities.is_editable(self._obj):
            return False

        return not AXObject.supports_value(self._obj)

    def get_braille_region(self) -> braille.Region | None:
        """Returns the braille region for this Zone."""

        if self._braille_region is not None:
            return self._braille_region

        region = braille.ReviewComponent(self._obj, self._string, 0, self)
        if self._can_cache_braille_region():
            self._braille_region = region
        return region

    def get_string(self) -> str:
        """Returns the string being displayed for this Zone."""

        return self._string

    def get_words(self) -> list[Word]:
        """Returns the list of Words in this Zone."""

        if self._words:
            return self._words

        # TODO - JD: For now, don't fake character and word extents.
        # The main goal is to improve reviewability.
        for i, word in enumerate(re.finditer(self.WORDS_RE, self._string)):
            word_obj = Word(self, word.start(), word.group())
            self._words.append(word_obj)
            key = (word_obj.get_start_offset(), word_obj.get_string())
            self._word_index_map[key] = i

        return self._words

    def get_word_at_index(self, index: int) -> Word | None:
        """Returns the Word at the specified index with respect to this Zone."""

        words = self.get_words()
        if 0 <= index < len(words):
            return words[index]

        return None

    def get_index_of_word(self, word: Word) -> int:
        """Returns the index of the specified Word with respect to this Zone."""

        # Ensure words are initialized
        self.get_words()
        key = (word.get_start_offset(), word.get_string())
        return self._word_index_map.get(key, -1)

    def get_object(self) -> Atspi.Accessible:
        """Returns the accessible object associated with this Zone."""

        return self._obj

    def get_rect(self) -> Atspi.Rect:
        """Returns the Atspi.Rect instance representing the extents of this Zone."""

        return self._rect

    def get_word_rect(self, start_offset: int, string: str) -> Atspi.Rect:
        """Returns the rectangle for a word within this zone."""

        cache_key = (start_offset, len(string))
        if cache_key in self._word_rect_cache:
            return self._word_rect_cache[cache_key]

        # TODO - JD: For now, don't fake character and word extents.
        # The main goal is to improve reviewability.
        rect = self._rect
        self._word_rect_cache[cache_key] = rect
        return rect

    def on_same_line(self, zone: Zone, pixel_delta: int = 5) -> bool:
        """Returns True if we treat this Zone and zone as being on one line."""

        if AXUtilities.is_scroll_bar(self._obj) or AXUtilities.is_scroll_bar(zone.get_object()):
            return self._obj == zone.get_object()

        this_rect = self._rect
        zone_rect = zone.get_rect()
        if this_rect.width == 0 and this_rect.height == 0:
            return zone_rect.y <= this_rect.y <= zone_rect.y + zone_rect.height

        if zone_rect.width == 0 and zone_rect.height == 0:
            return this_rect.y <= zone_rect.y <= this_rect.y + this_rect.height

        highest_bottom = min(this_rect.y + this_rect.height, zone_rect.y + zone_rect.height)
        lowest_top = max(this_rect.y, zone_rect.y)
        if lowest_top >= highest_bottom:
            return False

        this_middle = this_rect.y + this_rect.height / 2
        zone_middle = zone_rect.y + zone_rect.height / 2
        return abs(this_middle - zone_middle) <= pixel_delta

    def get_start_offset(self) -> int:
        """Returns the start offset of this Zone with respect to the accessible object."""

        return self._start_offset

    def get_word_at_offset(self, char_offset: int) -> tuple[Word | None, int]:
        """Returns the Word at the specified offset with respect to the accessible object."""

        msg = f"FLAT REVIEW: Searching for word at offset {char_offset}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        words = self.get_words()
        for word in words:
            tokens = ["FLAT REVIEW: Checking", word]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            offset = word.get_relative_offset(char_offset)
            if offset >= 0:
                return word, offset

        if len(self._string) == char_offset and words:
            last_word = words[-1]
            return last_word, len(last_word.get_string())

        return None, -1

    def has_caret(self) -> bool:
        """Returns True if this Zone contains the caret."""

        return False

    def word_with_caret(self) -> tuple[Word | None, int]:
        """Returns the Word and relative offset with the caret."""

        return None, -1


class TextZone(Zone):
    """A Zone whose purpose is to display text of an object."""

    def __init__(
        self,
        obj: Atspi.Accessible,
        start_offset: int,
        string: str,
        rect: Atspi.Rect,
    ) -> None:
        """Creates a new TextZone.

        Arguments:
        - obj: the Accessible associated with this Zone
        - string: the string being displayed for this Zone
        - rect: an Atspi.Rect instance representing the extents of this Zone
        """

        super().__init__(obj, string, rect)
        self._start_offset: int = start_offset
        self._word_rect_cache: dict[tuple[int, int], Atspi.Rect] = {}

    def __str__(self) -> str:
        text = self._string.replace("\n", "\\n")
        return f"TEXT ZONE: '{text}' {AXObject.get_role_name(self._obj)}"

    def _can_cache_braille_region(self) -> bool:
        """Returns True if we can cache the braille region for this Zone."""

        if AXUtilities.is_editable(self._obj) or AXUtilities.is_terminal(self._obj):
            return False

        return not AXUtilities.is_label(self._obj)

    def get_braille_region(self) -> braille.Region | None:
        """Returns the braille region for this Zone."""

        if self._braille_region is not None:
            return self._braille_region

        region = braille.ReviewText(self._obj, self._string, self._start_offset, self)
        if self._can_cache_braille_region():
            self._braille_region = region

        return region

    def get_string(self) -> str:
        """Returns the string being displayed for this Zone."""

        end_offset = self._start_offset + len(self._string)
        self._string = AXText.get_substring(self._obj, self._start_offset, end_offset)
        return self._string

    def get_words(self) -> list[Word]:
        """Returns the list of Words in this Zone."""

        words = []
        self._word_index_map.clear()  # Clear any existing mapping
        for i, word in enumerate(re.finditer(self.WORDS_RE, self.get_string())):
            start = word.start() + self._start_offset
            word_obj = Word(self, start, word.group())
            words.append(word_obj)
            key = (word_obj.get_start_offset(), word_obj.get_string())
            self._word_index_map[key] = i

        self._words = words
        return self._words

    def get_word_rect(self, start_offset: int, string: str) -> Atspi.Rect:
        """Returns the precise rectangle for a word within this TextZone."""

        cache_key = (start_offset, len(string))
        if cache_key in self._word_rect_cache:
            return self._word_rect_cache[cache_key]

        end_offset = start_offset + len(string)
        rect = AXText.get_range_rect(self._obj, start_offset, end_offset)
        self._word_rect_cache[cache_key] = rect
        return rect

    def has_caret(self) -> bool:
        """Returns True if this Zone contains the caret."""

        end_offset = self._start_offset + len(self._string)
        if self._start_offset <= AXText.get_caret_offset(self._obj) < end_offset:
            return True

        return end_offset == AXText.get_character_count(self._obj)

    def word_with_caret(self) -> tuple[Word | None, int]:
        """Returns the Word and relative offset with the caret."""

        if not self.has_caret():
            return None, -1

        return self.get_word_at_offset(AXText.get_caret_offset(self._obj))


class StateZone(Zone):
    """A Zone whose purpose is to display the state of an object."""

    def __init__(self, obj: Atspi.Accessible, rect: Atspi.Rect) -> None:
        super().__init__(obj, "", rect)

    def __str__(self) -> str:
        text = self.get_string().replace("\n", "\\n")
        return f"STATE ZONE: '{text}' {AXObject.get_role_name(self._obj)}"

    def _can_cache_braille_region(self) -> bool:
        """Returns True if we can cache the braille region for this Zone."""

        return False

    def get_braille_region(self) -> braille.Region | None:
        """Returns the braille region for this Zone."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return None

        result = script.get_braille_generator().get_state_indicator(self._obj)
        if result and result[0]:
            if isinstance(result[0], str):
                return braille.ReviewComponent(self._obj, result[0], 0, self)
            return result[0]
        return None

    def get_string(self) -> str:
        """Returns the string being displayed for this Zone."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return ""

        result = script.get_speech_generator().get_state_indicator(self._obj)
        return " ".join([r for r in result if isinstance(r, str)])


class ValueZone(Zone):
    """A Zone whose purpose is to display the value of an object."""

    def __init__(self, obj: Atspi.Accessible, rect: Atspi.Rect) -> None:
        super().__init__(obj, "", rect)

    def __str__(self) -> str:
        text = self.get_string().replace("\n", "\\n")
        return f"VALUE ZONE: '{text}' {AXObject.get_role_name(self._obj)}"

    def _can_cache_braille_region(self) -> bool:
        """Returns True if we can cache the braille region for this Zone."""

        return False

    def get_braille_region(self) -> braille.Region | None:
        """Returns the braille region for this Zone."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return None

        rolename = script.get_braille_generator().get_localized_role_name(self._obj)
        value = script.get_braille_generator().get_value(self._obj)
        if rolename and value:
            result = f"{rolename} {value[0]}"
            return braille.ReviewComponent(self._obj, result, 0, self)
        return None

    def get_string(self) -> str:
        """Returns the string being displayed for this Zone."""

        script = script_manager.get_manager().get_active_script()
        if script is None:
            return ""

        rolename = script.get_speech_generator().get_localized_role_name(self._obj)
        value = script.get_speech_generator().get_value(self._obj)
        result = ""
        if rolename and value:
            result = f"{rolename} {value[0]}"
        return result


class Line:
    """A Line is a single line across a window and is composed of Zones."""

    def __init__(self, line_number: int, zones: list[Zone]) -> None:
        """Creates a new Line, which is a horizontal region of text.

        Arguments:
        - line_number: the line number of this Line in the window
        - zones: the Zones that make up this line
        """

        self._line_number: int = line_number
        self._zones: list[Zone] = zones
        self._zone_index_map: dict[Zone, int] = {zone: i for i, zone in enumerate(zones)}

    def get_line_number(self) -> int:
        """Returns the index of this Line in the window."""

        return self._line_number

    def get_zones(self) -> list[Zone]:
        """Returns the list of Zones in this Line."""

        return self._zones

    def get_zone_at_index(self, index: int) -> Zone | None:
        """Returns the Zone at the specified index with respect to this Line."""

        if 0 <= index < len(self._zones):
            return self._zones[index]

        return None

    def get_index_of_zone(self, zone: Zone) -> int:
        """Returns the index of the specified Zone with respect to this Line."""

        return self._zone_index_map.get(zone, -1)

    def get_string(self) -> str:
        """Returns the string of this Line, which is the concatenation of all Zones."""

        return " ".join([zone.get_string() for zone in self._zones])

    def get_braille_regions(self) -> list[braille.Region]:
        """Returns a list of braille regions for this Line."""

        regions: list[braille.Region] = []
        for zone in self._zones:
            region = zone.get_braille_region()
            if region is not None:
                if regions:
                    regions.append(braille.Region(" "))
                regions.append(region)

        if braille_presenter.get_presenter().get_end_of_line_indicator_is_enabled():
            if regions:
                regions.append(braille.Region(" "))
            regions.append(braille.Region("$l"))

        return regions


class Context:
    """Contains the flat review regions for the current top-level object."""

    CHAR = 0
    WORD = 1
    ZONE = 2
    LINE = 3
    WINDOW = 4

    def __init__(self, script, root: Atspi.Accessible | None = None) -> None:
        """Create a new Context for script."""

        self._script = script
        self._zones: list[Zone] = []
        self._lines: list[Line] = []
        self._line_index: int = 0
        self._zone_index: int = 0
        self._word_index: int = 0
        self._char_index: int = 0
        self._focus_zone: Zone | None = None
        self._container: Atspi.Accessible | None = None
        self._object_to_zone_map: dict[Atspi.Accessible, list[Zone]] = {}
        self._focus_obj: Atspi.Accessible | None = focus_manager.get_manager().get_locus_of_focus()
        self._top_level: Atspi.Accessible | None = None
        self._rect: Atspi.Rect = Atspi.Rect()

        frame, dialog = script.utilities.frame_and_dialog(self._focus_obj)
        if root is not None:
            self._top_level = root
            tokens = ["FLAT REVIEW: Restricting flat review to", root]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        else:
            self._top_level = dialog or frame
        tokens = ["FLAT REVIEW: Frame:", frame, "Dialog:", dialog, ". Top level:", self._top_level]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        self._rect = AXComponent.get_rect(self._top_level)

        container = AXUtilities.find_ancestor(self._focus_obj, AXUtilities.is_menu)
        if container is None and AXUtilities.is_menu(self._focus_obj):
            container = self._focus_obj
        self._container = container or self._top_level

        self._zones = self._get_showing_zones(self._container)
        for zone in self._zones:
            obj = zone.get_object()
            if obj not in self._object_to_zone_map:
                self._object_to_zone_map[obj] = []
            self._object_to_zone_map[obj].append(zone)
        self._focus_zone = self._find_zone_with_object(self._focus_obj)

        self._lines = self._cluster_zones_by_line(self._zones)
        if not (self._lines and self._focus_zone):
            return

        for i, line in enumerate(self._lines):
            index = line.get_index_of_zone(self._focus_zone)
            if index < 0:
                continue

            self._line_index = i
            self._zone_index = index
            word, offset = self._focus_zone.word_with_caret()
            if word:
                self._word_index = self._focus_zone.get_index_of_word(word)
                self._char_index = offset
            break

        msg = (
            f"FLAT REVIEW: On line {self._line_index}, zone {self._zone_index} "
            f"word {self._word_index}, char {self._char_index}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _split_text_into_zones(
        self,
        obj: Atspi.Accessible,
        string: str,
        start_offset: int,
        cliprect: Atspi.Rect,
    ) -> list[TextZone]:
        """Returns a list of TextZones with embedded object characters removed."""

        zones: list[TextZone] = []
        ranges = [(*m.span(), m.group(0)) for m in re.finditer(r"[^\ufffc]+", string)]
        ranges = [(x[0] + start_offset, x[1] + start_offset, x[2]) for x in ranges]
        for start, end, substring in ranges:
            rect = AXText.get_range_rect(obj, start, end)
            intersection = AXUtilities.get_rect_intersection(rect, cliprect)
            if not AXUtilities.is_empty_rect(intersection):
                zones.append(TextZone(obj, start, substring, intersection))

        return zones

    def _get_zones_from_text(self, obj: Atspi.Accessible, cliprect: Atspi.Rect) -> list[TextZone]:
        """Returns a list of TextZones for the given object."""

        if not AXUtilities.has_presentable_text(obj):
            return []

        zones: list[TextZone] = []

        def _is_container(x: Atspi.Accessible) -> bool:
            return AXUtilities.is_scroll_pane(x) or AXUtilities.is_document(x)

        container = AXUtilities.find_ancestor(obj, _is_container)
        if container:
            rect = AXComponent.get_rect(container)
            intersection = AXUtilities.get_rect_intersection(rect, cliprect)
            if AXUtilities.is_same_rect(rect, intersection):
                tokens = ["FLAT REVIEW: Cliprect", cliprect, "->", rect, "from", container]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                cliprect = rect

        if AXObject.supports_editable_text(obj) and AXUtilities.is_single_line(obj):
            rect = AXComponent.get_rect(obj)
            return [TextZone(obj, 0, AXText.get_all_text(obj), rect)]

        tokens = ["FLAT REVIEW: Getting lines for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        lines = AXUtilities.get_visible_lines(obj, cliprect)
        # Optionally skip whitespace-only lines based on user preference.
        if not speech_presenter.get_presenter().get_speak_blank_lines():
            lines = [t for t in lines if t[0] and t[0].strip()]
        tokens = ["FLAT REVIEW:", len(lines), "lines found for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        for string, start_offset, _end_offset in lines:
            zones.extend(self._split_text_into_zones(obj, string, start_offset, cliprect))

        return zones

    def _get_zones_from_object(self, obj: Atspi.Accessible, cliprect: Atspi.Rect) -> list[Zone]:
        """Returns a list of Zones for the given object."""

        rect = AXComponent.get_rect(obj)
        zones: list[Zone] = list(self._get_zones_from_text(obj, cliprect))
        if not zones:
            if AXObject.supports_value(obj):
                zones.append(ValueZone(obj, rect))
            else:
                string = ""
                if not AXUtilities.is_table_row(obj):
                    string = self._script.get_speech_generator().get_name(obj, inFlatReview=True)
                if not string:
                    string = self._script.get_speech_generator().get_role_name(obj)
                if string:
                    zones.append(Zone(obj, string, rect))

        zone = StateZone(obj, rect)
        if zone.get_string():
            zones.insert(0, zone)

        return zones

    def set_current_to_zone_with_object(self, obj: Atspi.Accessible) -> bool:
        """Attempts to set the current zone to obj, if obj is in the current context."""

        tokens = [
            "FLAT REVIEW: Current",
            self.get_current_object(),
            f"line: {self._line_index}, zone: {self._zone_index},",
            f"word: {self._word_index}, char: {self._char_index})",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        zone = self._find_zone_with_object(obj)
        tokens = ["FLAT REVIEW: Zone with", obj, "is", zone]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if zone is None:
            return False

        for i, line in enumerate(self._lines):
            index = line.get_index_of_zone(zone)
            if index < 0:
                continue

            self._line_index = i
            self._zone_index = index
            word, offset = zone.word_with_caret()
            if word:
                self._word_index = zone.get_index_of_word(word)
                self._char_index = offset
            msg = "FLAT REVIEW: Updated current zone."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            break
        else:
            msg = "FLAT REVIEW: Failed to update current zone."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        tokens = [
            "FLAT REVIEW: Updated",
            self.get_current_object(),
            f"line: {self._line_index}, zone: {self._zone_index},",
            f"word: {self._word_index}, char: {self._char_index})",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    def _find_zone_with_object(self, obj: Atspi.Accessible | None) -> Zone | None:
        """Returns the existing zone which contains obj."""

        if obj is None:
            return None

        if matching_zones := self._object_to_zone_map.get(obj, []):
            for zone in matching_zones:
                if zone.has_caret():
                    return zone
            return matching_zones[0]

        # Items can be pruned from the flat review context. When this happens, usually the parent
        # or one of its children will still be in the context.
        if parent := AXObject.get_parent(obj):
            if parent_zones := self._object_to_zone_map.get(parent, []):
                return parent_zones[0]

        for child in AXObject.iter_children(obj):
            if child_zones := self._object_to_zone_map.get(child, []):
                return child_zones[0]

        return None

    def _get_showing_zones(
        self,
        root: Atspi.Accessible,
        boundingbox: Atspi.Rect | None = None,
    ) -> list[Zone]:
        """Returns an unsorted list of all the zones under root."""

        if boundingbox is None:
            boundingbox = self._rect

        if AXUtilities.is_menu(root):
            objs = []
            for child in AXObject.iter_children(root):
                objs.extend(AXUtilities.get_on_screen_objects(child))
        else:
            objs = AXUtilities.get_on_screen_objects(root, boundingbox)
        tokens = ["FLAT REVIEW:", len(objs), "on-screen objects found for", root]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        all_zones: list[Zone] = []
        obj_set = set(objs)
        for o in objs:
            if obj_set.intersection(AXUtilities.get_is_label_for(o)):
                continue

            zones = self._get_zones_from_object(o, boundingbox)
            if not zones:
                descendant = AXUtilities.active_descendant(o)
                if descendant:
                    zones = self._get_zones_from_object(descendant, boundingbox)
            all_zones.extend(zones)

        tokens = ["FLAT REVIEW:", len(all_zones), "zones found for", root]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return all_zones

    def _cluster_zones_by_line(self, zones: list[Zone]) -> list[Line]:
        """Returns a sorted list of Line clusters containing sorted Zones."""

        if not zones:
            return []

        zones.sort(key=lambda zone: zone.get_rect().y)
        line_clusters: list[list[Zone]] = []
        current_cluster = [zones[0]]
        for zone in zones[1:]:
            if zone.on_same_line(current_cluster[-1]):
                current_cluster.append(zone)
            else:
                current_cluster.sort(key=lambda z: z.get_rect().x)
                line_clusters.append(current_cluster)
                current_cluster = [zone]

        if current_cluster:
            current_cluster.sort(key=lambda z: z.get_rect().x)
            line_clusters.append(current_cluster)

        lines: list[Line] = []
        for line_index, zones_in_line in enumerate(line_clusters):
            line = Line(line_index, zones_in_line)
            lines.append(line)
            for zone in zones_in_line:
                zone.line = line

        tokens = ["FLAT REVIEW: Zones clustered into", len(lines), "lines"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return lines

    def get_current_line_string(self) -> str:
        """Returns the string of the current line."""

        if zone := self._get_current_zone():
            if zone.line:
                return zone.line.get_string()
        return ""

    def get_current_zone_string(self) -> str:
        """Returns the string of the current zone."""

        if zone := self._get_current_zone():
            return zone.get_string()
        return ""

    def get_current_word_string(self) -> str:
        """Returns the string of the current word."""

        zone = self._get_current_zone()
        if not zone:
            return ""

        if words := zone.get_words():
            return words[self._word_index].get_string()

        return zone.get_string()

    def get_current_character_string(self) -> str:
        """Returns the string of the current character."""

        zone = self._get_current_zone()
        if not zone:
            return ""

        words = zone.get_words()
        if not words:
            return ""

        word = words[self._word_index]
        if char := word.get_character_at_index(self._char_index):
            return char.get_string()
        return ""

    def _get_current_character_rect(self) -> Atspi.Rect:
        """Returns the extents of the current character."""

        zone = self._get_current_zone()
        if not zone:
            return Atspi.Rect()

        words = zone.get_words()
        if not words:
            return Atspi.Rect()

        word = words[self._word_index]
        if char := word.get_character_at_index(self._char_index):
            return char.get_rect()

        return Atspi.Rect()

    def get_current_location(self) -> tuple[int, int, int, int]:
        """Returns the location as a (lineIndex, zoneIndex, wordIndex, charIndex) tuple."""

        return self._line_index, self._zone_index, self._word_index, self._char_index

    def set_current_location(self, location: tuple[int, int, int, int]) -> None:
        """Sets the location to the specified (lineIndex, zoneIndex, wordIndex, charIndex)."""

        self._line_index, self._zone_index, self._word_index, self._char_index = location

    def set_current_zone(self, zone: Zone, offset_in_zone: int = 0) -> None:
        """Sets zone as the current zone."""

        if zone is None or zone.line is None:
            return

        self._line_index = zone.line.get_line_number()
        self._zone_index = self._lines[self._line_index].get_index_of_zone(zone)
        self._word_index = 0
        self._char_index = 0

        text_offset = zone.get_start_offset()
        word, character_offset = zone.get_word_at_offset(text_offset + offset_in_zone)
        if word:
            self._word_index = zone.get_index_of_word(word)
            self._char_index = character_offset

    def _get_current_zone(self) -> Zone | None:
        """Returns the current Zone."""

        if not (self._lines and 0 <= self._line_index < len(self._lines)):
            return None

        line = self._lines[self._line_index]
        zones = line.get_zones()
        if not (line and 0 <= self._zone_index < len(zones)):
            return None

        return zones[self._zone_index]

    def get_current_text_offset(self) -> int:
        """Returns the current text offset in the current object."""

        zone = self._get_current_zone()
        if zone is None:
            return -1
        words = zone.get_words()
        if not words:
            return -1
        word = words[self._word_index]
        if char := word.get_character_at_index(self._char_index):
            return char.get_start_offset()
        return -1

    def get_current_object(self) -> Atspi.Accessible | None:
        """Returns the current object."""

        zone = self._get_current_zone()
        if not zone:
            return None

        return zone.get_object()

    def get_current_braille_regions(self) -> tuple[list[braille.Region], braille.Region | None]:
        """Returns a (regions, focused-region) tuple."""

        if not self._lines:
            return [], None

        focused_region = None
        line = self._lines[self._line_index]
        regions = line.get_braille_regions()

        zone = self._get_current_zone()
        if zone is None:
            return regions, None

        focused_region = zone.get_braille_region()
        if focused_region is None:
            return regions, None

        focused_region.cursor_offset = 0
        if words := zone.get_words():
            focused_region.cursor_offset += words[0].get_start_offset() - zone.get_start_offset()
            for word_index in range(self._word_index):
                focused_region.cursor_offset += len(words[word_index].get_string())
        focused_region.cursor_offset += self._char_index
        # This is related to contracted braille.
        focused_region.reposition_cursor()
        return regions, focused_region

    def go_to_start_of(self, review_type: int = WINDOW) -> bool:
        """Returns True if moving to the start of the specified type succeeded."""

        before = [self._line_index, self._zone_index, self._word_index, self._char_index]
        self._char_index = 0
        self._word_index = 0
        if review_type == Context.WINDOW:
            self._zone_index = 0
            self._line_index = 0
        elif review_type == Context.LINE:
            self._zone_index = 0

        return before != [self._line_index, self._zone_index, self._word_index, self._char_index]

    def go_to_end_of(self, review_type: int = WINDOW) -> bool:
        """Returns True if moving to the end of the specified type succeeded."""

        if not self._lines:
            return False

        before = [self._line_index, self._zone_index, self._word_index, self._char_index]
        if review_type == Context.WINDOW:
            self._line_index = len(self._lines) - 1
        elif review_type == Context.LINE:
            self._zone_index = len(self._lines[self._line_index].get_zones()) - 1

        zone = self._get_current_zone()
        if zone is None:
            return False
        if words := zone.get_words():
            self._word_index = len(words) - 1
            word = words[self._word_index]
            chars = word.get_characters()
            self._char_index = max(len(chars) - 1, 0)
        else:
            self._word_index = 0
            self._char_index = 0

        return before != [self._line_index, self._zone_index, self._word_index, self._char_index]

    def go_previous_line(self, wrap: bool = False) -> bool:
        """Returns True if moving to the previous line succeeded."""

        if self._line_index > 0:
            self._line_index -= 1
            self._zone_index = 0
            self._word_index = 0
            self._char_index = 0
            return True

        if not wrap:
            return False

        self._line_index = max(0, len(self._lines) - 1)
        self._zone_index = 0
        self._word_index = 0
        self._char_index = 0
        return True

    def go_next_line(self, wrap: bool = False) -> bool:
        """Returns True if moving to the next line succeeded."""

        if self._line_index < (len(self._lines) - 1):
            self._line_index += 1
            self._zone_index = 0
            self._word_index = 0
            self._char_index = 0
            return True

        if not wrap:
            return False

        self._line_index = 0
        self._zone_index = 0
        self._word_index = 0
        self._char_index = 0
        return True

    def go_previous_zone(self) -> bool:
        """Returns True if moving to the previous zone succeeded."""

        if self._zone_index > 0:
            self._zone_index -= 1
            self._word_index = 0
            self._char_index = 0
            return True

        if not self.go_previous_line():
            return False

        zones = self._lines[self._line_index].get_zones()
        self._zone_index = max(len(zones) - 1, 0)
        self._word_index = 0
        self._char_index = 0
        return True

    def go_next_zone(self) -> bool:
        """Returns True if moving to the next zone succeeded."""

        zones = self._lines[self._line_index].get_zones()
        if self._zone_index < len(zones) - 1:
            self._zone_index += 1
            self._word_index = 0
            self._char_index = 0
            return True

        return self.go_next_line()

    def go_previous_word(self) -> bool:
        """Returns True if moving to the previous word succeeded."""

        if self._word_index > 0:
            self._word_index -= 1
            self._char_index = 0
            return True

        if not self.go_previous_zone():
            return False

        self.go_to_end_of(Context.ZONE)
        return True

    def go_next_word(self) -> bool:
        """Returns True if moving to the next word succeeded."""

        zone = self._get_current_zone()
        if zone is None:
            return False
        words = zone.get_words()
        if self._word_index < len(words) - 1:
            self._word_index += 1
            self._char_index = 0
            return True

        return self.go_next_zone()

    def go_previous_character(self) -> bool:
        """Returns True if moving to the previous character succeeded."""

        if self._char_index > 0:
            self._char_index -= 1
            return True

        if not self.go_previous_word():
            return False

        zone = self._get_current_zone()
        if zone is None:
            return False
        if words := zone.get_words():
            chars = words[self._word_index].get_characters()
            self._char_index = max(len(chars) - 1, 0)
        else:
            self._char_index = 0
        return True

    def go_next_character(self) -> bool:
        """Returns True if moving to the next character succeeded."""

        zone = self._get_current_zone()
        if zone is None:
            return False
        if words := zone.get_words():
            chars = words[self._word_index].get_characters()
            if self._char_index < (len(chars) - 1):
                self._char_index += 1
                return True

        return self.go_next_word()

    def go_up(self) -> bool:
        """Returns True if moving up succeeded."""

        rect = self._get_current_character_rect()
        target_x = rect.x + (rect.width / 2)
        if not self.go_previous_line():
            return False

        rect = self._get_current_character_rect()
        if rect.x + rect.width >= target_x:
            return True

        while self.go_next_character():
            rect = self._get_current_character_rect()
            if rect.x + rect.width >= target_x:
                break

        return True

    def go_down(self) -> bool:
        """Returns True if moving down succeeded."""

        rect = self._get_current_character_rect()
        target_x = rect.x + (rect.width / 2)
        if not self.go_next_line():
            return False

        rect = self._get_current_character_rect()
        if rect.x + rect.width >= target_x:
            return True

        while self.go_next_character():
            rect = self._get_current_character_rect()
            if rect.x + rect.width >= target_x:
                break

        return True
